import numpy as np
import neo
import elephant.spike_train_generation as stg
import elephant.statistics as stat
from copy import deepcopy
from MEArec.tools import *
import quantities as pq


class SpikeTrainGenerator:
    """
    Class for generation of spike trains called by the gen_recordings function.
    The list of parameters is in default_params/recordings_params.yaml (spiketrains field).

    Parameters
    ----------
    params :  dict
        Dictionary with parameters to simulate spiketrains. Default values can be retrieved with
        mr.get_default_recordings_params()['spiketrains']
    spiketrains : list of neo.SpikeTrain
        List of neo.SpikeTrain objects to instantiate a SpikeTrainGenerator with existing data
    verbose : bool
        If True, output is verbose
    """

    def __init__(self, params=None, spiketrains=None, seed=None, verbose=False):
        self._verbose = verbose
        self._has_spiketrains = False
        self.params = {}
        if params is None:
            if self._verbose:
                print("Using default parameters")
        if spiketrains is None:
            self.params = deepcopy(params)
            if seed is None:
                seed = np.random.randint(1000)
            if self._verbose:
                print('Spiketrains seed: ', seed)
            self.params['seed'] = seed
            np.random.seed(self.params['seed'])

            if 't_start' not in self.params.keys():
                params['t_start'] = 0
            self.params['t_start'] = params['t_start'] * pq.s
            if 'duration' not in self.params.keys():
                params['duration'] = 10
            self.params['t_stop'] = self.params['t_start'] + params['duration'] * pq.s
            if 'min_rate' not in self.params.keys():
                params['min_rate'] = 0.1
            self.params['min_rate'] = params['min_rate'] * pq.Hz
            if 'ref_per' not in self.params.keys():
                params['ref_per'] = 2
            self.params['ref_per'] = params['ref_per'] * pq.ms
            if 'process' not in self.params.keys():
                params['process'] = 'poisson'
            self.params['process'] = params['process']
            if 'gamma_shape' not in self.params.keys() and params['process'] == 'gamma':
                params['gamma_shape'] = 2
                self.params['gamma_shape'] = params['gamma_shape']

            if 'rates' in self.params.keys():  # all firing rates are provided
                self.params['rates'] = self.params['rates'] * pq.Hz
                self.n_neurons = len(self.params['rates'])
            else:
                rates = []
                types = []
                if 'f_exc' not in self.params.keys():
                    params['f_exc'] = 5
                self.params['f_exc'] = params['f_exc'] * pq.Hz
                if 'f_inh' not in self.params.keys():
                    params['f_inh'] = 15
                self.params['f_inh'] = params['f_inh'] * pq.Hz
                if 'st_exc' not in self.params.keys():
                    params['st_exc'] = 1
                self.params['st_exc'] = params['st_exc'] * pq.Hz
                if 'st_inh' not in self.params.keys():
                    params['st_inh'] = 3
                self.params['st_inh'] = params['st_inh'] * pq.Hz
                if 'n_exc' not in self.params.keys():
                    params['n_exc'] = 2
                self.params['n_exc'] = params['n_exc']
                if 'n_inh' not in self.params.keys():
                    params['n_inh'] = 1
                self.params['n_inh'] = params['n_inh']

                for exc in np.arange(self.params['n_exc']):
                    rate = self.params['st_exc'] * np.random.randn() + self.params['f_exc']
                    if rate < self.params['min_rate']:
                        rate = self.params['min_rate']
                    rates.append(rate)
                    types.append('e')
                for inh in np.arange(self.params['n_inh']):
                    rate = self.params['st_inh'] * np.random.randn() + self.params['f_inh']
                    if rate < self.params['min_rate']:
                        rate = self.params['min_rate']
                    rates.append(rate)
                    types.append('i')
                self.params['rates'] = rates
                self.params['types'] = types
                self.n_neurons = len(self.params['rates'])

            self.info = params
            self.spiketrains = False
        else:
            self.spiketrains = spiketrains
            self.info = {}
            self._has_spiketrains = True
            if params is not None:
                self.params = deepcopy(params)

    def set_spiketrain(self, idx, spiketrain):
        """
        Sets spike train idx to new spiketrain.

        Parameters
        ----------
        idx : int
            Index of spike train to set
        spiketrain : neo.SpikeTrain
            New spike train

        """
        self.spiketrains[idx] = spiketrain

    def generate_spikes(self):
        """
        Generate spike trains based on default_params of the SpikeTrainGenerator class.
        self.spiketrains contains the newly generated spike trains
        """
        if not self._has_spiketrains:
            self.spiketrains = []
            idx = 0
            for n in np.arange(self.n_neurons):
                rate = self.params['rates'][n]
                if self.params['process'] == 'poisson':
                    st = stg.homogeneous_poisson_process(rate,
                                                         self.params['t_start'], self.params['t_stop'])
                elif self.params['process'] == 'gamma':
                    st = stg.homogeneous_gamma_process(self.params['gamma_shape'], rate,
                                                       self.params['t_start'], self.params['t_stop'])
                self.spiketrains.append(st)
                self.spiketrains[-1].annotate(fr=rate)
                if 'n_exc' in self.params.keys() and 'n_inh' in self.params.keys():
                    if idx < self.params['n_exc']:
                        self.spiketrains[-1].annotate(cell_type='E')
                    else:
                        self.spiketrains[-1].annotate(cell_type='I')
                idx += 1

            # check consistency and remove spikes below refractory period
            for idx, st in enumerate(self.spiketrains):
                isi = stat.isi(st)
                idx_remove = np.where(isi < self.params['ref_per'])[0]
                spikes_to_remove = len(idx_remove)
                unit = st.times.units

                while spikes_to_remove > 0:
                    new_times = np.delete(st.times, idx_remove[0]) * unit
                    st = neo.SpikeTrain(new_times, t_start=self.params['t_start'], t_stop=self.params['t_stop'])
                    isi = stat.isi(st)
                    idx_remove = np.where(isi < self.params['ref_per'])[0]
                    spikes_to_remove = len(idx_remove)

                st.annotations = self.spiketrains[idx].annotations
                self.set_spiketrain(idx, st)
        else:
            print("SpikeTrainGenerator initialized with existing spike trains!")

    def add_synchrony(self, idxs, rate=0.05, time_jitt=1 * pq.ms, verbose=False):
        """
        Adds synchronous spikes between pairs of spike trains at a certain rate.

        Parameters
        ----------
        idxs : list or array
            Spike train indexes to add synchrony to
        rate : float
            Rate of added synchrony spike to spike train idxs[1] for each spike of idxs[0]
        time_jitt : quantity
            Maximum time jittering between added spikes
        verbose : bool
            If True output is verbose

        Returns
        -------
        sync_rate : float
            New synchrony rate
        fr1 : quantity
            Firing rate spike train 1
        fr2 : quantity
            Firing rate spike train 2

        """
        idx1 = idxs[0]
        idx2 = idxs[1]
        st1 = self.spiketrains[idx1]
        st2 = self.spiketrains[idx2]
        times1 = st1.times
        times2 = st2.times
        t_start = st2.t_start
        t_stop = st2.t_stop
        unit = times2.units
        added_spikes_t1 = 0
        added_spikes_t2 = 0

        all_times_shuffle = np.concatenate((times1, times2))
        all_times_shuffle = all_times_shuffle[np.random.permutation(len(all_times_shuffle))] * unit

        sync_rate = compute_sync_rate(st1, st2, time_jitt)
        if sync_rate < rate:
            for t in all_times_shuffle:
                if sync_rate <= rate:
                    # check time difference
                    if t in times1:
                        t_diff = np.abs(t.rescale(pq.ms).magnitude - times2.rescale(pq.ms).magnitude)
                        if np.all(t_diff > self.params['ref_per']):
                            t1_jitt = time_jitt.rescale(unit).magnitude * np.random.rand(1) + t.rescale(
                                unit).magnitude - \
                                      (time_jitt.rescale(unit) / 2).magnitude
                            if t1_jitt < t_stop:
                                times2 = np.sort(np.concatenate((np.array(times2), np.array(t1_jitt))))
                                times2 = times2 * unit
                                st2 = neo.SpikeTrain(times2, t_start=t_start, t_stop=t_stop)
                                added_spikes_t1 += 1
                    elif t in times2:
                        t_diff = np.abs(t.rescale(pq.ms).magnitude - times1.rescale(pq.ms).magnitude)
                        if np.all(t_diff > self.params['ref_per']):
                            t2_jitt = time_jitt.rescale(unit).magnitude * np.random.rand(1) + t.rescale(
                                unit).magnitude - \
                                      (time_jitt.rescale(unit) / 2).magnitude
                            if t2_jitt < t_stop:
                                times1 = np.sort(np.concatenate((np.array(times1), np.array(t2_jitt))))
                                times1 = times1 * unit
                                st1 = neo.SpikeTrain(times1, t_start=t_start, t_stop=t_stop)
                                added_spikes_t2 += 1
                    sync_rate = compute_sync_rate(st1, st2, time_jitt)
                else:
                    break
            if verbose:
                print("Added", added_spikes_t1, "spikes to spike train", idxs[0],
                      "and", added_spikes_t2, "spikes to spike train", idxs[1], 'Sync rate:', sync_rate)
        else:
            spiketrains = [st1, st2]
            annotate_overlapping_spikes(spiketrains)
            max_overlaps = np.floor(rate * (len(times1) + len(times2)))
            curr_overlaps = np.floor(sync_rate * (len(times1) + len(times2)))
            remove_overlaps = int(curr_overlaps - max_overlaps)
            if curr_overlaps > max_overlaps:
                st1_ovrl_idx = np.where(spiketrains[0].annotations['overlap'] == 'TO')[0]
                st2_ovrl_idx = np.where(spiketrains[1].annotations['overlap'] == 'TO')[0]
                perm = np.random.permutation(len(st1_ovrl_idx))[:remove_overlaps]
                st1_ovrl_idx = st1_ovrl_idx[perm]
                st2_ovrl_idx = st2_ovrl_idx[perm]
                idx_rm_1 = st1_ovrl_idx[:remove_overlaps // 2]
                idx_rm_2 = st2_ovrl_idx[remove_overlaps // 2:]
                times1 = np.delete(st1.times, idx_rm_1)
                times1 = times1 * unit
                times2 = np.delete(st2.times, idx_rm_2)
                times2 = times2 * unit
                st1 = neo.SpikeTrain(times1, t_start=t_start, t_stop=t_stop)
                st2 = neo.SpikeTrain(times2, t_start=t_start, t_stop=t_stop)
                sync_rate = compute_sync_rate(st1, st2, time_jitt)
                if verbose:
                    print("Removed", len(idx_rm_1), "spikes from spike train", idxs[0],
                          "and", len(idx_rm_2), "spikes from spike train", idxs[1], 'Sync rate:', sync_rate)

        st1.annotations = self.spiketrains[idx1].annotations
        st2.annotations = self.spiketrains[idx2].annotations
        self.set_spiketrain(idx1, st1)
        self.set_spiketrain(idx2, st2)

        fr1 = len(st1.times) / st1.t_stop
        fr2 = len(st2.times) / st2.t_stop

        return sync_rate, fr1, fr2


import numpy as np
import time
from copy import deepcopy
from MEArec.tools import *
import MEAutility as mu
import shutil
import yaml
import os
from os.path import join
from distutils.version import StrictVersion
from pathlib import Path
from joblib import Parallel, delayed

if StrictVersion(yaml.__version__) >= StrictVersion('5.0.0'):
    use_loader = True
else:
    use_loader = False


def simulate_cell_templates(i, simulate_script, tot, cell_model,
                            model_folder, intraonly, params, verbose):
    print("Starting ", i + 1)
    print('\n\n', cell_model, i + 1, '/', tot, '\n\n')
    os.system('python %s %s %s %s %s' \
              % (simulate_script, join(model_folder, cell_model), intraonly, params, verbose))
    print("Exiting ", i + 1)


class TemplateGenerator:
    """
    Class for generation of templates called by the gen_templates function.
    The list of parameters is in default_params/templates_params.yaml.

    Parameters
    ----------
    cell_models_folder : str
        Path to folder containing Blue Brain Project cell models
    templates_folder : str
        Path to output template folder (if not in params)
    temp_dict :  dict
        Dictionary to instantiate TemplateGenerator with existing data. It contains the following fields:
          - templates : float (n_templates, n_electrodes, n_timepoints)
          - locations : float (n_templates, 3)
          - rotations : float (n_templates, 3)
          - celltypes : str (n_templates)
    info :  dict
        Info dictionary to instantiate TemplateGenerator with existing data. It contains the following fields:
          - params : dict with template generation parameters
          - electrodes : dict with probe info (from MEAutility.return_mea_info('probe-name'))
    params : dict
        Dictionary with parameters to simulate templates. Default values can be retrieved with
        mr.get_default_template_params()
    intraonly : bool
        If True, only intracellular simulations are performed
    parallel : bool
        If True, cell models are run in parallel
    recompile: bool
        If True, cell models are recompiled (suggested if new models are added)
    n_jobs: int
        If None, all cpus are used
    delete_tmp : bool
        If True, temporary files are removed
    verbose : bool
        If True, output is verbose
    """

    def __init__(self, cell_models_folder=None, templates_folder=None, temp_dict=None, info=None,
                 params=None, intraonly=False, parallel=True, recompile=False, n_jobs=None, delete_tmp=True,
                 verbose=False):
        self._verbose = verbose
        if temp_dict is not None and info is not None:
            if 'templates' in temp_dict.keys():
                self.templates = temp_dict['templates']
            if 'locations' in temp_dict.keys():
                self.locations = temp_dict['locations']
            if 'rotations' in temp_dict.keys():
                self.rotations = temp_dict['rotations']
            if 'celltypes' in temp_dict.keys():
                self.celltypes = temp_dict['celltypes']
            self.info = info
            self.params = deepcopy(info)
        else:
            if cell_models_folder is None:
                raise AttributeError("Specify cell folder!")
            if params is None:
                if self._verbose:
                    print("Using default parameters")
                self.params = {}
            else:
                self.params = deepcopy(params)
            self.cell_model_folder = cell_models_folder
            self.n_jobs = n_jobs
            self.templates_folder = templates_folder
            self.simulation_params = {'intraonly': intraonly, 'parallel': parallel, 'delete_tmp': delete_tmp,
                                      'recompile': recompile}

    def generate_templates(self):
        """
        Generate templates.
        """
        cell_models_folder = self.cell_model_folder
        templates_folder = self.templates_folder
        intraonly = self.simulation_params['intraonly']
        parallel = self.simulation_params['parallel']
        recompile = self.simulation_params['recompile']
        delete_tmp = self.simulation_params['delete_tmp']

        if os.path.isdir(cell_models_folder):
            cell_models = [f for f in os.listdir(join(cell_models_folder)) if 'mods' not in f]
            if len(cell_models) == 0:
                raise AttributeError(cell_models_folder, ' contains no cell models!')
        else:
            raise NotADirectoryError('Cell models folder: does not exist!')

        this_dir, this_filename = os.path.split(__file__)
        simulate_script = str(Path(this_dir).parent / 'simulate_cells.py')
        self.params['cell_models_folder'] = cell_models_folder
        self.params['templates_folder'] = templates_folder

        # Compile NEURON models (nrnivmodl)
        if not os.path.isdir(join(cell_models_folder, 'mods')) or recompile:
            if self._verbose:
                print('Compiling NEURON models')
            os.system('python %s compile %s' % (simulate_script, cell_models_folder))

        if 'sim_time' not in self.params.keys():
            self.params['sim_time'] = 1
        if 'target_spikes' not in self.params.keys():
            self.params['target_spikes'] = [3, 50]
        if 'cut_out' not in self.params.keys():
            self.params['cut_out'] = [2, 5]
        if 'dt' not in self.params.keys():
            self.params['dt'] = 2 ** -5
        if 'delay' not in self.params.keys():
            self.params['delay'] = 10
        if 'weights' not in self.params.keys():
            self.params['weights'] = [0.25, 1.75]

        if 'rot' not in self.params.keys():
            self.params['rot'] = 'physrot'
        if 'probe' not in self.params.keys():
            available_mea = mu.return_mea_list()
            probe = available_mea[np.random.randint(len(available_mea))]
            if self._verbose:
                print("Probe randomly set to: %s" % probe)
            self.params['probe'] = probe
        if 'ncontacts' not in self.params.keys():
            self.params['ncontacts'] = 1
        if 'overhang' not in self.params.keys():
            self.params['overhang'] = 1
        if 'xlim' not in self.params.keys():
            self.params['xlim'] = [10, 80]
        if 'ylim' not in self.params.keys():
            self.params['ylim'] = None
        if 'zlim' not in self.params.keys():
            self.params['zlim'] = None
        if 'offset' not in self.params.keys():
            self.params['offset'] = 0
        if 'det_thresh' not in self.params.keys():
            self.params['det_thresh'] = 30
        if 'n' not in self.params.keys():
            self.params['n'] = 50
        if 'min_amp' not in self.params.keys():
            self.params['min_amp'] = 30
        if 'seed' not in self.params.keys():
            self.params['seed'] = np.random.randint(1, 10000)
        elif self.params['seed'] is None:
            self.params['seed'] = np.random.randint(1, 10000)
        if templates_folder is None:
            self.params['templates_folder'] = os.getcwd()
            templates_folder = self.params['templates_folder']
        else:
            self.params['templates_folder'] = templates_folder
        if 'drifting' not in self.params.keys():
            self.params['drifting'] = False
        if 'max_drift' not in self.params.keys():
            self.params['max_drift'] = 100
        if 'min_drift' not in self.params.keys():
            self.params['min_drift'] = 30
        if 'drift_steps' not in self.params.keys():
            self.params['drift_steps'] = 10
        if 'drift_xlim' not in self.params.keys():
            self.params['drift_xlim'] = [-10, 10]
        if 'drift_ylim' not in self.params.keys():
            self.params['drift_ylim'] = [-10, 10]
        if 'drift_zlim' not in self.params.keys():
            self.params['drift_zlim'] = [20, 80]

        rot = self.params['rot']
        n = self.params['n']
        probe = self.params['probe']

        tmp_params_path = 'tmp_params_path'
        with open(tmp_params_path, 'w') as f:
            yaml.dump(self.params, f)

        # Simulate neurons and EAP for different cell models separately
        if parallel and self.n_jobs not in (0, 1):
            start_time = time.time()
            tot = len(cell_models)
            if self.n_jobs is None:
                n_jobs = tot
            else:
                n_jobs = self.n_jobs

            if self._verbose:
                print('Running with', n_jobs, 'jobs')

            Parallel(n_jobs=n_jobs)(delayed(simulate_cell_templates)(i, simulate_script, tot, cell_model,
                                                                     cell_models_folder, intraonly, tmp_params_path,
                                                                     self._verbose, )
                                    for i, cell_model in enumerate(cell_models))
            print('\n\n\nSimulation time: ', time.time() - start_time, '\n\n\n')
        else:
            # TODO try without subprocess
            start_time = time.time()
            for numb, cell_model in enumerate(cell_models):
                if self._verbose:
                    print('\n\n', cell_model, numb + 1, '/', len(cell_models), '\n\n')
                os.system('python %s %s %s %s %s' \
                          % (simulate_script, join(cell_models_folder, cell_model), intraonly, tmp_params_path,
                             self._verbose))
            print('\n\n\nSimulation time: ', time.time() - start_time, '\n\n\n')

        tmp_folder = join(templates_folder, rot, 'tmp_%d_%s' % (n, probe))

        if not Path(tmp_folder).is_dir():
            raise FileNotFoundError(f'{tmp_folder} not found. Something went wrong in the template generation phase.')

        templates, locations, rotations, celltypes = load_tmp_eap(tmp_folder)
        if delete_tmp:
            shutil.rmtree(tmp_folder)
            os.remove(tmp_params_path)

        self.info = {}

        self.templates = templates
        self.locations = locations
        self.rotations = rotations
        self.celltypes = celltypes
        self.info['params'] = self.params
        self.info['electrodes'] = mu.return_mea_info(probe)


import numpy as np
import time
from copy import deepcopy
from MEArec.tools import (select_templates, find_overlapping_templates, get_binary_cat,
                          resample_templates, jitter_templates, pad_templates, get_templates_features,
                          compute_modulation, annotate_overlapping_spikes, extract_wf)

from .recgensteps import (chunk_convolution, chunk_uncorrelated_noise,
                          chunk_distance_correlated_noise, chunk_apply_filter)

import random
import string
import MEAutility as mu
import yaml
import os
import shutil
import h5py
import quantities as pq
from distutils.version import StrictVersion
import tempfile
from pathlib import Path

from joblib import Parallel, delayed

from MEArec.generators import SpikeTrainGenerator

if StrictVersion(yaml.__version__) >= StrictVersion('5.0.0'):
    use_loader = True
else:
    use_loader = False


class RecordingGenerator:
    """
    Class for generation of recordings called by the gen_recordings function.
    The list of parameters is in default_params/recordings_params.yaml.

    Parameters
    ----------
    spgen : SpikeTrainGenerator
        SpikeTrainGenerator object containing spike trains
    tempgen : TemplateGenerator
        TemplateGenerator object containing templates
    params : dict
        Dictionary with parameters to simulate recordings. Default values can be retrieved with
        mr.get_default_recording_params()
    rec_dict :  dict
        Dictionary to instantiate RecordingGenerator with existing data. It contains the following fields:
          - recordings : float (n_electrodes, n_samples)
          - spiketrains : list of neo.SpikeTrains (n_spiketrains)
          - templates : float (n_spiketrains, 3)
          - template_locations : float (n_spiketrains, 3)
          - template_rotations : float (n_spiketrains, 3)
          - template_celltypes : str (n_spiketrains)
          - channel_positions : float (n_electrodes, 3)
          - timestamps : float (n_samples)
          - voltage_peaks : float (n_spiketrains, n_electrodes)
          - spike_traces : float (n_spiketrains, n_samples)
    info :  dict
        Info dictionary to instantiate RecordingGenerator with existing data. Same fields as 'params'
    """

    def __init__(self, spgen=None, tempgen=None, params=None, rec_dict=None, info=None):
        self._verbose = False
        self._verbose_1 = False
        self._verbose_2 = False

        if rec_dict is not None and info is not None:
            if 'recordings' in rec_dict.keys():
                self.recordings = rec_dict['recordings']
            else:
                self.recordings = np.array([])
            if 'spiketrains' in rec_dict.keys():
                self.spiketrains = deepcopy(rec_dict['spiketrains'])
            else:
                self.spiketrains = np.array([])
            if 'templates' in rec_dict.keys():
                self.templates = rec_dict['templates']
            else:
                self.templates = np.array([])
            if 'template_locations' in rec_dict.keys():
                self.template_locations = rec_dict['template_locations']
            else:
                self.template_locations = np.array([])
            if 'template_rotations' in rec_dict.keys():
                self.template_rotations = rec_dict['template_rotations']
            else:
                self.template_rotations = np.array([])
            if 'template_celltypes' in rec_dict.keys():
                self.template_celltypes = rec_dict['template_celltypes']
            else:
                self.template_celltypes = np.array([])
            if 'channel_positions' in rec_dict.keys():
                self.channel_positions = rec_dict['channel_positions']
            else:
                self.channel_positions = np.array([])
            if 'timestamps' in rec_dict.keys():
                self.timestamps = rec_dict['timestamps']
            else:
                self.timestamps = np.array([])
            if 'voltage_peaks' in rec_dict.keys():
                self.voltage_peaks = rec_dict['voltage_peaks']
            else:
                self.voltage_peaks = np.array([])
            if 'spike_traces' in rec_dict.keys():
                self.spike_traces = rec_dict['spike_traces']
            else:
                self.spike_traces = np.array([])
            self.info = deepcopy(info)
            self.params = deepcopy(info)
            if len(self.spiketrains) > 0:
                if 'spiketrains' in self.info:
                    self.spgen = SpikeTrainGenerator(spiketrains=self.spiketrains, params=self.info['spiketrains'])
                else:
                    self.spgen = SpikeTrainGenerator(spiketrains=self.spiketrains, params={'custom': True})
            self.tempgen = None
            if isinstance(self.recordings, np.memmap):
                self.tmp_mode = 'memmap'
            else:
                self.tmp_mode = None

        else:
            if spgen is None or tempgen is None:
                raise AttributeError("Specify SpikeTrainGenerator and TemplateGenerator objects!")
            if params is None:
                params = {'spiketrains': {},
                          'celltypes': {},
                          'templates': {},
                          'recordings': {},
                          'seeds': {}}
            self.params = deepcopy(params)
            self.spgen = spgen
            self.tempgen = tempgen
            self.tmp_mode = None

        self.overlapping = []
        # temp file that should remove on delete
        self._to_remove_on_delete = []
        self.tmp_folder = None
        self._is_tmp_folder_local = False

    def __del__(self):
        self.recordings = None
        self.spike_traces = None

        if not self._is_tmp_folder_local:
            if self.tmp_folder is not None:
                try:
                    shutil.rmtree(self.tmp_folder)
                    if self._verbose >= 1:
                        print('Deleted', self.tmp_folder)
                except Exception as e:
                    if self._verbose >= 1:
                        print('Impossible to delete temp file:', self.tmp_folder, 'Error', e)
        else:
            for fname in self._to_remove_on_delete:
                try:
                    os.remove(fname)
                    if self._verbose >= 1:
                        print('Deleted', fname)
                except Exception as e:
                    if self._verbose >= 1:
                        print('Impossible to delete temp file:', fname, 'Error', e)

    def generate_recordings(self, tmp_mode=None, tmp_folder=None, verbose=None, n_jobs=0):
        """
        Generates recordings

        Parameters
        ----------
        tmp_mode : None, 'memmap'
            Use temporary file h5 memmap or None
            None is no temporary file and then use memory.
        tmp_folder: str or Path
            In case of tmp files, you can specify the folder.
            If None, then it is automatic using tempfile.mkdtemp()
        verbose: bool or int
            Determines the level of verbose. If 1 or True, low-level, if 2 high level, if False, not verbose
        n_jobs: int if >1 then use joblib to execute chunk in parallel else in loop
        """

        self.tmp_mode = tmp_mode
        self.tmp_folder = tmp_folder
        self.n_jobs = n_jobs

        if tmp_mode is not None:
            tmp_prefix = ''.join([random.choice(string.ascii_letters) for i in range(5)]) + '_'
        if self.tmp_mode is not None:
            if self.tmp_folder is None:
                self.tmp_folder = Path(tempfile.mkdtemp())
                self._is_tmp_folder_local = False
            else:
                self.tmp_folder = Path(self.tmp_folder)
                self._is_tmp_folder_local = True
        else:
            self._is_tmp_folder_local = False

        self._verbose = verbose
        if self._verbose is not None and isinstance(self._verbose, bool) or isinstance(self._verbose, int):
            verbose_1 = self._verbose >= 1
            verbose_2 = self._verbose >= 2
        elif isinstance(verbose, bool):
            if self._verbose:
                verbose_1 = True
                verbose_2 = False
            else:
                verbose_1 = False
                verbose_2 = False
        else:  # None
            verbose_1 = False
            verbose_2 = False
        self._verbose_1 = verbose_1
        self._verbose_2 = verbose_2

        params = deepcopy(self.params)
        temp_params = self.params['templates']
        rec_params = self.params['recordings']
        st_params = self.params['spiketrains']
        seeds = self.params['seeds']

        if 'cell_types' in self.params.keys():
            celltype_params = self.params['cell_types']
        else:
            celltype_params = {}

        tempgen = self.tempgen
        spgen = self.spgen
        if tempgen is not None:
            eaps = tempgen.templates
            locs = tempgen.locations
            rots = tempgen.rotations
            celltypes = tempgen.celltypes
            temp_info = tempgen.info
            cut_outs = temp_info['params']['cut_out']
        else:
            temp_info = None
            cut_outs = self.params['templates']['cut_out']

        spiketrains = spgen.spiketrains
        n_neurons = len(spiketrains)

        if len(spiketrains) > 0:
            duration = spiketrains[0].t_stop - spiketrains[0].t_start
            only_noise = False
        else:
            if verbose_1:
                print('No spike trains provided: only simulating noise')
            only_noise = True
            duration = st_params['duration'] * pq.s

        if 'fs' not in rec_params.keys() and temp_info is not None:
            # when computed from templates fs is in kHz
            params['recordings']['fs'] = 1. / temp_info['params']['dt']
            fs = (params['recordings']['fs'] * pq.kHz).rescale('Hz')
            spike_fs = fs
        elif params['recordings']['fs'] is None and temp_info is not None:
            params['recordings']['fs'] = 1. / temp_info['params']['dt']
            fs = (params['recordings']['fs'] * pq.kHz).rescale('Hz')
            spike_fs = fs
        else:
            # In the rec_params fs is in Hz
            fs = params['recordings']['fs'] * pq.Hz
            if temp_info is not None:
                spike_fs = (1. / temp_info['params']['dt'] * pq.kHz).rescale('Hz')
            else:
                spike_fs = fs

        if 'dtype' not in rec_params.keys():
            params['recordings']['dtype'] = 'float32'
        elif rec_params['dtype'] is None:
            params['recordings']['dtype'] = 'float32'
        else:
            params['recordings']['dtype'] = rec_params['dtype']
        dtype = params['recordings']['dtype']

        if 'noise_mode' not in rec_params.keys():
            params['recordings']['noise_mode'] = 'uncorrelated'
        noise_mode = params['recordings']['noise_mode']

        if 'noise_color' not in rec_params.keys():
            params['recordings']['noise_color'] = False
        noise_color = params['recordings']['noise_color']

        if 'sync_rate' not in rec_params.keys():
            params['recordings']['sync_rate'] = None
        sync_rate = params['recordings']['sync_rate']

        if 'sync_jitt' not in rec_params.keys():
            params['recordings']['sync_jitt'] = 1
        sync_jitt = params['recordings']['sync_jitt'] * pq.ms

        if noise_mode == 'distance-correlated':
            if 'noise_half_distance' not in rec_params.keys():
                params['recordings']['noise_half_distance'] = 30
            half_dist = params['recordings']['noise_half_distance']

        if noise_mode == 'far-neurons':
            if 'far_neurons_n' not in rec_params.keys():
                params['recordings']['far_neurons_n'] = 300
            far_neurons_n = params['recordings']['far_neurons_n']
            if 'far_neurons_max_amp' not in rec_params.keys():
                params['recordings']['far_neurons_max_amp'] = 20
            far_neurons_max_amp = params['recordings']['far_neurons_max_amp']
            if 'far_neurons_noise_floor' not in rec_params.keys():
                params['recordings']['far_neurons_noise_floor'] = 0.5
            far_neurons_noise_floor = params['recordings']['far_neurons_noise_floor']
            if 'far_neurons_exc_inh_ratio' not in rec_params.keys():
                params['recordings']['far_neurons_exc_inh_ratio'] = 0.8
            far_neurons_exc_inh_ratio = params['recordings']['far_neurons_exc_inh_ratio']

        if noise_color:
            if 'color_peak' not in rec_params.keys():
                params['recordings']['color_peak'] = 500
            color_peak = params['recordings']['color_peak']
            if 'color_q' not in rec_params.keys():
                params['recordings']['color_q'] = 1
            color_q = params['recordings']['color_q']
            if 'color_noise_floor' not in rec_params.keys():
                params['recordings']['color_noise_floor'] = 1
            color_noise_floor = params['recordings']['color_noise_floor']
        else:
            color_peak, color_q, color_noise_floor = None, None, None

        if 'noise_level' not in rec_params.keys():
            params['recordings']['noise_level'] = 10
        noise_level = params['recordings']['noise_level']

        if verbose_1:
            print('Noise Level ', noise_level)

        if 'filter' not in rec_params.keys():
            params['recordings']['filter'] = True
        filter = params['recordings']['filter']

        if 'filter_cutoff' not in rec_params.keys():
            params['recordings']['filter_cutoff'] = [300., 6000.]
        cutoff = params['recordings']['filter_cutoff'] * pq.Hz

        if 'filter_order' not in rec_params.keys():
            params['recordings']['filter_order'] = 3
        order = params['recordings']['filter_order']

        if 'modulation' not in rec_params.keys():
            params['recordings']['modulation'] = 'electrode'
        elif params['recordings']['modulation'] not in ['none', 'electrode', 'template']:
            raise Exception("'modulation' can be 'none', 'template', or 'electrode'")
        modulation = params['recordings']['modulation']

        if 'bursting' not in rec_params.keys():
            params['recordings']['bursting'] = False
        bursting = params['recordings']['bursting']

        if 'shape_mod' not in rec_params.keys():
            params['recordings']['shape_mod'] = False
        shape_mod = params['recordings']['shape_mod']

        if bursting:
            if 'exp_decay' not in rec_params.keys():
                params['recordings']['exp_decay'] = 0.2
            exp_decay = params['recordings']['exp_decay']
            if 'n_burst_spikes' not in rec_params.keys():
                params['recordings']['n_burst_spikes'] = 10
            n_burst_spikes = params['recordings']['n_burst_spikes']
            if 'max_burst_duration' not in rec_params.keys():
                params['recordings']['max_burst_duration'] = 100
            max_burst_duration = params['recordings']['max_burst_duration'] * pq.ms

            if rec_params['n_bursting'] is None:
                n_bursting = n_neurons
            else:
                n_bursting = rec_params['n_bursting']

            if shape_mod:
                if 'shape_stretch' not in rec_params.keys():
                    params['recordings']['shape_stretch'] = 30
                shape_stretch = params['recordings']['shape_stretch']
                if verbose_1:
                    print('Bursting with modulation sigmoid: ', shape_stretch)
            else:
                shape_stretch = None
        else:
            exp_decay = None
            n_burst_spikes = None
            max_burst_duration = None
            shape_stretch = None
            n_bursting = None

        chunk_duration = params['recordings'].get('chunk_duration', 0) * pq.s
        if chunk_duration == 0 * pq.s:
            chunk_duration = duration

        if 'mrand' not in rec_params.keys():
            params['recordings']['mrand'] = 1
        mrand = params['recordings']['mrand']

        if 'sdrand' not in rec_params.keys():
            params['recordings']['sdrand'] = 0.05
        sdrand = params['recordings']['sdrand']

        if 'overlap' not in rec_params.keys():
            params['recordings']['overlap'] = False
        overlap = params['recordings']['overlap']

        if 'extract_waveforms' not in rec_params.keys():
            params['recordings']['extract_waveforms'] = False
        extract_waveforms = params['recordings']['extract_waveforms']

        if 'xlim' not in temp_params.keys():
            params['templates']['xlim'] = None
        x_lim = params['templates']['xlim']

        if 'ylim' not in temp_params.keys():
            params['templates']['ylim'] = None
        y_lim = params['templates']['ylim']

        if 'zlim' not in temp_params.keys():
            params['templates']['zlim'] = None
        z_lim = params['templates']['zlim']

        if 'n_overlap_pairs' not in temp_params.keys():
            params['templates']['n_overlap_pairs'] = None
        n_overlap_pairs = params['templates']['n_overlap_pairs']

        if 'min_amp' not in temp_params.keys():
            params['templates']['min_amp'] = 50
        min_amp = params['templates']['min_amp']

        if 'max_amp' not in temp_params.keys():
            params['templates']['max_amp'] = np.inf
        max_amp = params['templates']['max_amp']

        if 'min_dist' not in temp_params.keys():
            params['templates']['min_dist'] = 25
        min_dist = params['templates']['min_dist']

        if 'overlap_threshold' not in temp_params.keys():
            params['templates']['overlap_threshold'] = 0.8
        overlap_threshold = params['templates']['overlap_threshold']

        if 'pad_len' not in temp_params.keys():
            params['templates']['pad_len'] = [3., 3.]
        pad_len = params['templates']['pad_len']

        if 'n_jitters' not in temp_params.keys():
            params['templates']['n_jitters'] = 10
        n_jitters = params['templates']['n_jitters']

        if 'upsample' not in temp_params.keys():
            params['templates']['upsample'] = 8
        upsample = params['templates']['upsample']

        if 'drifting' not in rec_params.keys():
            params['recordings']['drifting'] = False
        drifting = params['recordings']['drifting']

        # set seeds
        if 'templates' not in seeds.keys():
            temp_seed = np.random.randint(1, 1000)
        elif seeds['templates'] is None:
            temp_seed = np.random.randint(1, 1000)
        else:
            temp_seed = seeds['templates']

        if 'convolution' not in seeds.keys():
            conv_seed = np.random.randint(1, 1000)
        elif seeds['convolution'] is None:
            conv_seed = np.random.randint(1, 1000)
        else:
            conv_seed = seeds['convolution']

        if 'noise' not in seeds.keys():
            noise_seed = np.random.randint(1, 1000)
        elif seeds['noise'] is None:
            noise_seed = np.random.randint(1, 1000)
        else:
            noise_seed = seeds['noise']

        if drifting:
            if temp_info is not None:
                assert temp_info['params']['drifting'], "For generating drifting recordings, templates must be drifting"
            else:
                if params['n_jitters'] == 1:
                    assert len(self.templates.shape) == 4
                else:
                    assert len(self.templates.shape) == 5
            preferred_dir = np.array(rec_params['preferred_dir'])
            preferred_dir = preferred_dir / np.linalg.norm(preferred_dir)
            angle_tol = rec_params['angle_tol']
            drift_velocity = rec_params['slow_drift_velocity']
            fast_drift_period = rec_params['fast_drift_period'] * pq.s
            fast_drift_max_jump = rec_params['fast_drift_max_jump']
            fast_drift_min_jump = rec_params['fast_drift_min_jump']
            drift_mode = rec_params['drift_mode']
            t_start_drift = rec_params['t_start_drift'] * pq.s
            if rec_params['n_drifting'] is None:
                n_drifting = n_neurons
            else:
                n_drifting = rec_params['n_drifting']
            if 'fast' in drift_mode:
                if chunk_duration > 0 and chunk_duration != duration:
                    print('Disabling chunking for fast drifts')
                    chunk_duration = duration
        else:
            # if drifting templates, but not recordings, consider initial template
            if temp_info is not None:
                if temp_info['params']['drifting']:
                    eaps = eaps[:, 0]
                    locs = locs[:, 0]
            elif len(self.templates.shape) == 5:
                self.templates = self.templates[:, 0]
                self.template_locs = self.template_locs[:, 0]
            preferred_dir = None
            angle_tol = None
            drift_velocity = None
            fast_drift_period = None
            fast_drift_max_jump = None
            fast_drift_min_jump = None
            t_start_drift = None
            n_drifting = None
            drift_mode = None

        # load MEA info
        if temp_info is not None:
            mea = mu.return_mea(info=temp_info['electrodes'])
            params['electrodes'] = temp_info['electrodes']
        else:
            mea = mu.return_mea(info=self.params['electrodes'])
            params['electrodes'] = self.params['electrodes']
        mea_pos = mea.positions
        n_elec = mea_pos.shape[0]
        n_samples = int(duration.rescale('s').magnitude * fs.rescale('Hz').magnitude)

        params['recordings'].update({'duration': float(duration.magnitude),
                                     'fs': float(fs.rescale('Hz').magnitude),
                                     'n_neurons': n_neurons})
        params['templates'].update({'cut_out': cut_outs})

        # create buffer h5/memmap/memmory
        if self.tmp_mode == 'memmap':
            tmp_path_0 = self.tmp_folder / (tmp_prefix + "mearec_tmp_file_recordings.raw")
            recordings = np.memmap(tmp_path_0, shape=(n_samples, n_elec), dtype=dtype, mode='w+')
            recordings[:] = 0
            # recordings = recordings.transpose()
            tmp_path_1 = self.tmp_folder / (tmp_prefix + "mearec_tmp_file_spike_traces.raw")
            if not only_noise:
                spike_traces = np.memmap(tmp_path_1, shape=(n_samples, n_neurons), dtype=dtype, mode='w+')
                spike_traces[:] = 0
                # spike_traces = spike_traces.transpose()
            # file names for templates
            tmp_templates_pad = self.tmp_folder / (tmp_prefix + "templates_pad.raw")
            tmp_templates_rs = self.tmp_folder / (tmp_prefix + "templates_resample.raw")
            tmp_templates_jit = self.tmp_folder / (tmp_prefix + "templates_jitter.raw")
            self._to_remove_on_delete.extend([tmp_path_0, tmp_path_1,
                                              tmp_templates_pad, tmp_templates_rs, tmp_templates_jit])
        else:
            recordings = np.zeros((n_samples, n_elec), dtype=dtype)
            spike_traces = np.zeros((n_samples, n_neurons), dtype=dtype)
            tmp_templates_pad = None
            tmp_templates_rs = None
            tmp_templates_jit = None

        timestamps = np.arange(recordings.shape[0]) / fs

        #######################
        # Step 1: convolution #
        #######################
        if only_noise:
            spiketrains = np.array([])
            voltage_peaks = np.array([])
            spike_traces = np.array([])
            templates = np.array([])
            template_locs = np.array([])
            template_rots = np.array([])
            template_celltypes = np.array([])
            overlapping = np.array([])
        else:
            if tempgen is not None:
                if celltype_params is not None:
                    if 'excitatory' in celltype_params.keys() and 'inhibitory' in celltype_params.keys():
                        exc_categories = celltype_params['excitatory']
                        inh_categories = celltype_params['inhibitory']
                        bin_cat = get_binary_cat(celltypes, exc_categories, inh_categories)
                    else:
                        bin_cat = np.array(['U'] * len(celltypes))
                else:
                    bin_cat = np.array(['U'] * len(celltypes))

                if 'cell_type' in spiketrains[0].annotations.keys():
                    n_exc = [st.annotations['cell_type'] for st in spiketrains].count('E')
                    n_inh = n_neurons - n_exc
                    st_types = np.array([st.annotations['cell_type'] for st in spiketrains])
                elif 'rates' in st_params.keys():
                    assert st_params['types'] is not None, "If 'rates' are provided as spiketrains parameters, " \
                                                           "corresponding 'types' ('E'-'I') must be provided"
                    n_exc = st_params['types'].count('E')
                    n_inh = st_params['types'].count('I')
                    st_types = np.array(st_params['types'])
                else:
                    if self._verbose:
                        print('Setting random number of excitatory and inhibitory neurons as cell_type info is missing')
                    n_exc = np.random.randint(n_neurons)
                    n_inh = n_neurons - n_exc
                    st_types = np.array(['E'] * n_exc + ['I'] * n_inh)

                e_idx = np.where(st_types == 'E')
                i_idx = np.where(st_types == 'I')
                if len(e_idx) > 0 and len(i_idx) > 0:
                    if not np.all([[e < i for e in e_idx[0]] for i in i_idx[0]]):
                        if verbose_1:
                            print('Re-arranging spike trains: Excitatory first, Inhibitory last')
                        order = np.argsort(st_types)
                        new_spiketrains = []
                        for idx in order:
                            new_spiketrains.append(spiketrains[idx])
                        spgen.spiketrains = new_spiketrains
                        spiketrains = new_spiketrains

                if verbose_1:
                    print('Templates selection seed: ', temp_seed)
                np.random.seed(temp_seed)

                if drifting:
                    drift_directions = np.array([(p[-1] - p[0]) / np.linalg.norm(p[-1] - p[0]) for p in locs])
                    drift_velocity_ums = drift_velocity / 60.
                    velocity_vector = drift_velocity_ums * preferred_dir
                    if verbose_1:
                        print('Drift mode: ', drift_mode)
                        if 'slow' in drift_mode:
                            print('Slow drift velocity', drift_velocity, 'um/min')
                        if 'fast' in drift_mode:
                            print('Fast drift period', fast_drift_period)
                            print('Fast drift max jump',
                                  fast_drift_max_jump)  # 'Fast drift min jump', fast_drift_min_jump)
                    n_elec = eaps.shape[2]
                else:
                    drift_directions = None
                    preferred_dir = None
                    velocity_vector = None
                    n_elec = eaps.shape[1]

                if n_neurons > 100 or drifting:
                    parallel_templates = True
                else:
                    parallel_templates = False

                if verbose_1:
                    print('Selecting cells')
                idxs_cells, selected_cat = select_templates(locs, eaps, bin_cat, n_exc, n_inh, x_lim=x_lim, y_lim=y_lim,
                                                            z_lim=z_lim, min_amp=min_amp, max_amp=max_amp,
                                                            min_dist=min_dist, drifting=drifting,
                                                            drift_dir=drift_directions,
                                                            preferred_dir=preferred_dir, angle_tol=angle_tol,
                                                            n_overlap_pairs=n_overlap_pairs,
                                                            overlap_threshold=overlap_threshold,
                                                            verbose=verbose_2)

                if not np.any('U' in  selected_cat):
                    assert selected_cat.count('E') == n_exc and selected_cat.count('I') == n_inh
                    # Reorder templates according to E-I types
                    reordered_idx_cells = np.array(idxs_cells)[np.argsort(selected_cat)]
                else:
                    reordered_idx_cells = idxs_cells

                template_celltypes = celltypes[reordered_idx_cells]
                template_locs = np.array(locs)[reordered_idx_cells]
                template_rots = np.array(rots)[reordered_idx_cells]
                template_bin = np.array(bin_cat)[reordered_idx_cells]
                templates = np.array(eaps)[reordered_idx_cells]

                # find overlapping templates
                overlapping = find_overlapping_templates(templates, thresh=overlap_threshold)

                # peak images
                voltage_peaks = []
                for tem in templates:
                    dt = 1. / fs.magnitude
                    if not drifting:
                        feat = get_templates_features(tem, ['neg'], dt=dt)
                    else:
                        feat = get_templates_features(tem[0], ['neg'], dt=dt)
                    voltage_peaks.append(-np.squeeze(feat['neg']))
                voltage_peaks = np.array(voltage_peaks)

                # pad templates
                pad_samples = [int((pp * fs.rescale('kHz')).magnitude) for pp in pad_len]
                if verbose_1:
                    print('Padding template edges')
                t_pad = time.time()
                templates_pad = pad_templates(templates, pad_samples, drifting, dtype, verbose_2,
                                              tmp_file=tmp_templates_pad, parallel=parallel_templates)

                if verbose_1:
                    print('Elapsed pad time:', time.time() - t_pad)

                # resample spikes
                t_rs = time.time()
                up = fs
                down = spike_fs
                spike_duration_pad = templates_pad.shape[-1]
                if up != down:
                    n_resample = int(spike_duration_pad * (up / down))
                    templates_rs = resample_templates(templates_pad, n_resample, up, down, drifting, dtype,
                                                      verbose_2, tmp_file=tmp_templates_rs,
                                                      parallel=parallel_templates)
                    if verbose_1:
                        print('Elapsed resample time:', time.time() - t_rs)
                else:
                    templates_rs = templates_pad

                if verbose_1:
                    print('Creating time jittering')
                jitter = 1. / fs
                t_j = time.time()
                templates = jitter_templates(templates_rs, upsample, fs, n_jitters, jitter, drifting, dtype,
                                             verbose_2, tmp_file=tmp_templates_jit, parallel=parallel_templates)
                if verbose_1:
                    print('Elapsed jitter time:', time.time() - t_j)

                # find cut out samples for convolution after padding and resampling
                pre_peak_fraction = (pad_len[0] + cut_outs[0]) / (np.sum(pad_len) + np.sum(cut_outs))
                samples_pre_peak = int(pre_peak_fraction * templates.shape[-1])
                samples_post_peak = templates.shape[-1] - samples_pre_peak
                cut_outs_samples = [samples_pre_peak, samples_post_peak]

                # delete temporary preprocessed templates
                del templates_rs, templates_pad
            else:
                templates = self.templates
                pre_peak_fraction = (pad_len[0] + cut_outs[0]) / (np.sum(pad_len) + np.sum(cut_outs))
                samples_pre_peak = int(pre_peak_fraction * templates.shape[-1])
                samples_post_peak = templates.shape[-1] - samples_pre_peak
                cut_outs_samples = [samples_pre_peak, samples_post_peak]
                template_locs = self.template_locations
                template_rots = self.template_rotations
                template_celltypes = self.template_celltypes
                if celltype_params is not None:
                    if 'excitatory' in celltype_params.keys() and 'inhibitory' in celltype_params.keys():
                        exc_categories = celltype_params['excitatory']
                        inh_categories = celltype_params['inhibitory']
                        template_bin = get_binary_cat(template_celltypes, exc_categories, inh_categories)
                    else:
                        template_bin = np.array(['U'] * len(celltypes))
                else:
                    template_bin = np.array(['U'] * len(celltypes))
                voltage_peaks = self.voltage_peaks
                overlapping = np.array([])
                if not drifting:
                    velocity_vector = None
                else:
                    drift_velocity_ums = drift_velocity / 60.
                    velocity_vector = drift_velocity_ums * preferred_dir
                    if verbose_1:
                        print('Drift velocity vector: ', velocity_vector)

            if sync_rate is not None:
                if verbose_1:
                    print('Modifying synchrony of spatially overlapping spikes')
                if verbose_1:
                    print('Overlapping templates: ', overlapping)
                for over in overlapping:
                    if verbose_1:
                        print('Overlapping pair: ', over)
                    spgen.add_synchrony(over, rate=sync_rate, verbose=verbose_2, time_jitt=sync_jitt)
                    # annotate new firing rates
                    fr1 = len(spgen.spiketrains[over[0]].times) / spgen.spiketrains[over[0]].t_stop
                    fr2 = len(spgen.spiketrains[over[1]].times) / spgen.spiketrains[over[1]].t_stop
                    spgen.spiketrains[over[0]].annotate(fr=fr1)
                    spgen.spiketrains[over[1]].annotate(fr=fr2)
            self.overlapping = overlapping

            # find SNR and annotate
            if verbose_1:
                print('Computing spike train SNR')
            for t_i, temp in enumerate(templates):
                min_peak = np.min(temp)
                snr = np.abs(min_peak / float(noise_level))
                spiketrains[t_i].annotate(snr=snr)

            if verbose_1:
                print('Adding spiketrain annotations')
            for i, st in enumerate(spiketrains):
                st.annotate(bintype=template_bin[i], mtype=template_celltypes[i], soma_position=template_locs[i])

            if overlap:
                annotate_overlapping_spikes(spiketrains, overlapping_pairs=overlapping, verbose=verbose_2)

            if verbose_1:
                print('Convolution seed: ', conv_seed)
            np.random.seed(conv_seed)

            amp_mod = []
            cons_spikes = []

            # modulated convolution
            if drifting:
                drifting_units = np.random.permutation(n_neurons)[:n_drifting]
            else:
                drifting_units = []
            if bursting:
                bursting_units = np.random.permutation(n_neurons)[:n_bursting]
            else:
                bursting_units = []

            if modulation == 'template':
                if verbose_1:
                    print('Template modulation')
                for i_s, st in enumerate(spiketrains):
                    if bursting and i_s in bursting_units:
                        if verbose_1:
                            print('Bursting unit: ', i_s)
                        amp, cons = compute_modulation(st, sdrand=sdrand,
                                                       n_spikes=n_burst_spikes, exp=exp_decay,
                                                       max_burst_duration=max_burst_duration)

                        amp_mod.append(amp)
                        cons_spikes.append(cons)
                        st.annotate(bursting=True)
                    else:
                        amp, cons = compute_modulation(st, mrand=mrand, sdrand=sdrand,
                                                       n_spikes=0)
                        amp_mod.append(amp)
                        cons_spikes.append(cons)
            elif modulation == 'electrode':
                if verbose_1:
                    print('Electrode modulaton')
                for i_s, st in enumerate(spiketrains):
                    if bursting and i_s in bursting_units:
                        if verbose_1:
                            print('Bursting unit: ', i_s)
                        amp, cons = compute_modulation(st, n_el=n_elec, mrand=mrand, sdrand=sdrand,
                                                       n_spikes=n_burst_spikes, exp=exp_decay,
                                                       max_burst_duration=max_burst_duration)
                        amp_mod.append(amp)

                        cons_spikes.append(cons)
                        st.annotate(bursting=True)
                    else:
                        amp, cons = compute_modulation(st, n_el=n_elec, mrand=mrand, sdrand=sdrand,
                                                       n_spikes=0)
                        amp_mod.append(amp)
                        cons_spikes.append(cons)

            spike_idxs = []
            for st in spiketrains:
                spike_idxs.append((st.times * fs).magnitude.astype('int'))

            # divide in chunks
            chunk_indexes = make_chunk_indexes(duration, chunk_duration, fs)
            seed_list_conv = [np.random.randint(1000) for i in np.arange(len(chunk_indexes))]

            pad_samples_conv = templates.shape[-1]
            # call the loop on chunks
            args = (spike_idxs, pad_samples_conv, modulation, drifting, drift_mode, drifting_units, templates,
                    cut_outs_samples, template_locs, velocity_vector, fast_drift_period,
                    fast_drift_min_jump, fast_drift_max_jump, t_start_drift, fs, verbose_2,
                    amp_mod, bursting_units, shape_mod, shape_stretch,
                    True, voltage_peaks, dtype, seed_list_conv,)
            assignment_dict = {
                'recordings': recordings,
                'spike_traces': spike_traces}
            output_list = run_several_chunks(chunk_convolution, chunk_indexes, fs, timestamps, args,
                                             self.n_jobs, self.tmp_mode, self.tmp_folder, assignment_dict)

            # if drift then propagate annoations to spikestrains
            for st in np.arange(n_neurons):
                if drifting and st in drifting_units:
                    spiketrains[st].annotate(drifting=True)
                    template_idxs = np.array([], dtype='int')
                    for out in output_list:
                        template_idxs = np.concatenate((template_idxs, out['template_idxs'][st]))
                    assert len(template_idxs) == len(spiketrains[st])
                    spiketrains[st].annotate(template_idxs=template_idxs)

        #################
        # Step 2: noise #
        #################
        if verbose_1:
            print('Adding noise')
            print('Noise seed: ', noise_seed)

        np.random.seed(noise_seed)

        if noise_level == 0:
            if verbose_1:
                print('Noise level is set to 0')
        else:
            # divide in chunks
            chunk_indexes = make_chunk_indexes(duration, chunk_duration, fs)
            seed_list_noise = [np.random.randint(1000) for i in np.arange(len(chunk_indexes))]

            if self.tmp_mode == 'memmap':
                tmp_path_noise = self.tmp_folder / (tmp_prefix + "mearec_tmp_noise_file.raw")
                additive_noise = np.memmap(tmp_path_noise, shape=(n_samples, n_elec), dtype=dtype, mode='w+')
                # additive_noise = additive_noise.transpose()
                self._to_remove_on_delete.append(tmp_path_noise)
            else:
                additive_noise = np.zeros((n_samples, n_elec), dtype=dtype)

            if noise_mode == 'uncorrelated':
                func = chunk_uncorrelated_noise
                args = (n_elec, noise_level, noise_color, color_peak, color_q, color_noise_floor,
                        fs.rescale('Hz').magnitude, dtype, seed_list_noise,)
                assignment_dict = {'additive_noise': additive_noise}

                run_several_chunks(func, chunk_indexes, fs, timestamps, args,
                                   self.n_jobs, self.tmp_mode, self.tmp_folder, assignment_dict)

            elif noise_mode == 'distance-correlated':
                cov_dist = np.zeros((n_elec, n_elec))
                for i, el in enumerate(mea.positions):
                    for j, p in enumerate(mea.positions):
                        if i != j:
                            cov_dist[i, j] = (0.5 * half_dist) / np.linalg.norm(el - p)
                        else:
                            cov_dist[i, j] = 1

                func = chunk_distance_correlated_noise
                args = (noise_level, cov_dist, n_elec, noise_color, color_peak, color_q, color_noise_floor,
                        fs.rescale('Hz').magnitude, dtype, seed_list_noise,)
                assignment_dict = {'additive_noise': additive_noise}

                run_several_chunks(func, chunk_indexes, fs, timestamps, args,
                                   self.n_jobs, self.tmp_mode, self.tmp_folder, assignment_dict)

            elif noise_mode == 'far-neurons':
                if self.tmp_mode == 'memmap':
                    # file names for templates
                    tmp_templates_noise_pad = self.tmp_folder / (tmp_prefix + "templates_noise_pad.raw")
                    tmp_templates_noise_rs = self.tmp_folder / (tmp_prefix + "templates_noise_resample.raw")
                    self._to_remove_on_delete.extend([tmp_templates_noise_pad, tmp_templates_noise_rs])
                else:
                    tmp_templates_noise_pad = None
                    tmp_templates_noise_rs = None
                idxs_cells, selected_cat = select_templates(locs, eaps, bin_cat=None, n_exc=far_neurons_n, n_inh=0,
                                                            x_lim=x_lim, y_lim=y_lim, z_lim=z_lim, min_amp=0,
                                                            max_amp=far_neurons_max_amp, min_dist=1,
                                                            verbose=False)
                idxs_cells = sorted(idxs_cells)
                templates_noise = eaps[idxs_cells]
                template_noise_locs = locs[idxs_cells]
                if drifting:
                    templates_noise = templates_noise[:, 0]

                # pad spikes
                pad_samples = [int((pp * fs.rescale('kHz')).magnitude) for pp in pad_len]
                if verbose_1:
                    print('Padding noisy template edges')
                t_pad = time.time()
                templates_noise_pad = pad_templates(templates_noise, pad_samples, drifting, dtype, verbose_2,
                                                    tmp_file=tmp_templates_noise_pad, parallel=True)
                if verbose_1:
                    print('Elapsed pad time:', time.time() - t_pad)

                # resample templates
                t_rs = time.time()
                up = fs
                down = spike_fs
                spike_duration_pad = templates_noise_pad.shape[-1]
                if up != down:
                    n_resample = int(spike_duration_pad * (up / down))
                    templates_noise = resample_templates(templates_noise_pad, n_resample, up, down,
                                                         drifting, dtype, verbose_2,
                                                         tmp_file=tmp_templates_noise_rs)
                    if verbose_1:
                        print('Elapsed resample time:', time.time() - t_rs)
                else:
                    templates_noise = templates_noise_pad

                # find cut out samples for convolution after padding and resampling
                pre_peak_fraction = (pad_len[0] + cut_outs[0]) / (np.sum(pad_len) + np.sum(cut_outs))
                samples_pre_peak = int(pre_peak_fraction * templates.shape[-1])
                samples_post_peak = templates_noise.shape[-1] - samples_pre_peak
                cut_outs_samples = [samples_pre_peak, samples_post_peak]

                del templates_noise_pad

                # create noisy spiketrains
                if verbose_1:
                    print('Generating noisy spike trains')
                noisy_spiketrains_params = params['spiketrains']
                noisy_spiketrains_params['n_exc'] = int(far_neurons_n * far_neurons_exc_inh_ratio)
                noisy_spiketrains_params['n_inh'] = far_neurons_n - noisy_spiketrains_params['n_exc']
                noisy_spiketrains_params['seed'] = noise_seed
                spgen_noise = SpikeTrainGenerator(params=noisy_spiketrains_params)
                spgen_noise.generate_spikes()
                spiketrains_noise = spgen_noise.spiketrains

                spike_idxs_noise = []
                for st in spiketrains_noise:
                    spike_idxs_noise.append((st.times * fs).magnitude.astype('int'))

                if verbose_1:
                    print('Convolving noisy spike trains')
                templates_noise = templates_noise.reshape((templates_noise.shape[0], 1, templates_noise.shape[1],
                                                           templates_noise.shape[2]))

                # call the loop on chunks
                args = (spike_idxs_noise, 0, 'none', False, None, None, templates_noise,
                        cut_outs_samples, template_noise_locs, None, None, None, None, None, None,
                        verbose_2, None, None, False, None, False, None, dtype, seed_list_noise,)
                assignment_dict = {'recordings': additive_noise}
                run_several_chunks(chunk_convolution, chunk_indexes, fs, timestamps, args,
                                   self.n_jobs, self.tmp_mode, self.tmp_folder, assignment_dict)

                # removing mean
                for i, m in enumerate(np.mean(additive_noise, axis=0)):
                    additive_noise[:, i] -= m
                # adding noise floor
                for i, s in enumerate(np.std(additive_noise, axis=0)):
                    additive_noise[:, i] += far_neurons_noise_floor * s * \
                                                np.random.randn(additive_noise.shape[0])
                # scaling noise
                noise_scale = noise_level / np.std(additive_noise, axis=0)
                if verbose_1:
                    print('Scaling to reach desired level')
                for i, n in enumerate(noise_scale):
                    additive_noise[:, i] *= n

            # Add it to recordings
            recordings += additive_noise

        ##################
        # Step 3: filter #
        ##################
        if filter:
            if verbose_1:
                print('Filtering')
                if cutoff.size == 1:
                    print('High-pass cutoff', cutoff)
                elif cutoff.size == 2:
                    print('Band-pass cutoff', cutoff)

            chunk_indexes = make_chunk_indexes(duration, chunk_duration, fs)

            # compute pad samples as 3 times the low-cutoff period
            if cutoff.size == 1:
                pad_samples_filt = 3 * int((1. / cutoff * fs).magnitude)
            elif cutoff.size == 2:
                pad_samples_filt = 3 * int((1. / cutoff[0] * fs).magnitude)

            # call the loop on chunks
            args = (recordings, pad_samples_filt, cutoff, order, fs, dtype,)
            assignment_dict = {
                'filtered_chunk': recordings,
            }
            # Done in loop (as before) : this cannot be done in parralel because of bug transpose in joblib!!!!!!!!!!!!!
            run_several_chunks(chunk_apply_filter, chunk_indexes, fs, timestamps, args,
                               self.n_jobs, self.tmp_mode, self.tmp_folder, assignment_dict)

        # assign class variables
        params['templates']['overlapping'] = np.array(overlapping)
        self.recordings = recordings.transpose()
        self.timestamps = timestamps
        self.channel_positions = mea_pos
        self.templates = np.squeeze(templates)
        self.template_locations = template_locs
        self.template_rotations = template_rots
        self.template_celltypes = template_celltypes
        self.spiketrains = spiketrains
        self.voltage_peaks = voltage_peaks
        self.spike_traces = spike_traces.transpose()
        self.info = params

        #############################
        # Step 4: extract waveforms #
        #############################
        if not only_noise:
            if extract_waveforms:
                if verbose_1:
                    print('Extracting spike waveforms')
                self.extract_waveforms()

    def annotate_overlapping_spikes(self, parallel=True):
        """
        Annnotate spike trains with overlapping information.

        parallel : bool
            If True, spike trains are annotated in parallel
        """
        if self.info['templates']['overlapping'] is None or len(self.info['templates']['overlapping']) == 0:
            if self._verbose_1:
                print('Finding overlapping spikes')
            if len(self.templates.shape) == 3:
                templates = self.templates
            elif len(self.templates.shape) == 4:
                # drifting + no jitt or no drifting + jitt
                templates = self.templates[:, 0]
            elif len(self.templates.shape) == 5:
                # drifting + jitt
                templates = self.templates[:, 0, 0]
            self.overlapping = find_overlapping_templates(templates,
                                                          thresh=self.info['templates']['overlap_threshold'])
            print('Overlapping templates: ', self.overlapping)
            self.info['templates']['overlapping'] = self.overlapping
        annotate_overlapping_spikes(self.spiketrains, overlapping_pairs=self.overlapping, parallel=parallel)

    def extract_waveforms(self, cut_out=[0.5, 2]):
        """
        Extract waveforms from spike trains and recordings.

        Parameters
        ----------
        cut_out : float or list
            Ms before and after peak to cut out. If float the cut is symmetric.
        """
        fs = self.info['recordings']['fs'] * pq.Hz
        extract_wf(self.spiketrains, self.recordings, fs=fs, cut_out=cut_out)

    def extract_templates(self, cut_out=[0.5, 2], recompute=False):
        """
        Extract templates from spike trains.

        Parameters
        ----------
        cut_out : float or list
            Ms before and after peak to cut out. If float the cut is symmetric.
        recompute :  bool
            If True, templates are recomputed from extracted waveforms
        """
        fs = self.info['recordings']['fs'] * pq.Hz

        if not len(self.spiketrains) == 0:
            if self.spiketrains[0].waveforms is None:
                extract_wf(self.spiketrains, self.recordings, fs=fs, cut_out=cut_out)

        if self.tempgen is None or len(self.templates) == 0 or not recompute:
            wfs = [st.waveforms for st in self.spiketrains]
            templates = np.array([np.mean(wf, axis=0) for wf in wfs])
            if np.array(cut_out).size == 1:
                cut_out = [cut_out, cut_out]
            self.info['templates']['cut_out'] = cut_out
            self.info['templates']['pad_len'] = [0, 0]
            self.templates = templates[:, np.newaxis]
        else:
            raise Exception("templates are already computed. Use the 'recompute' argument to compute them from "
                            "extracted waveforms")


def make_chunk_indexes(total_duration, chunk_duration, fs):
    """
    Construct chunks list.
    Return a list of (start, stop) indexes.
    """
    fs_float = fs.rescale('Hz').magnitude
    chunk_size = int(chunk_duration.rescale('s').magnitude * fs_float)
    total_length = int(total_duration.rescale('s').magnitude * fs_float)

    if chunk_size == 0:
        chunk_indexes = [(0, total_length), ]
    else:
        n = int(np.floor(total_length / chunk_size))
        chunk_indexes = [(i * chunk_size, (i + 1) * chunk_size) for i in range(n)]
        if (total_length % chunk_size) > 0:
            chunk_indexes.append((n * chunk_size, total_length))

    return chunk_indexes


def run_several_chunks(func, chunk_indexes, fs, timestamps, args, n_jobs, tmp_mode, tmp_folder, assignment_dict):
    """
    Run a function on a list of chunks.
    
    this can be done in loop if n_jobs=1 (or 0)
    or in paralell if n_jobs>1
    
    The function can return
    
    
    """

    # create task list
    arg_tasks = []
    karg_tasks = []
    for ch, (i_start, i_stop) in enumerate(chunk_indexes):

        chunk_start = (i_start / fs).rescale('s')

        arg_task = (ch, i_start, i_stop, chunk_start,) + args
        arg_tasks.append(arg_task)

        karg_task = dict(assignment_dict=assignment_dict, tmp_mode=tmp_mode)
        karg_tasks.append(karg_task)

    # run chunks
    if n_jobs in (0, 1):
        # simple loop
        output_list = []
        for ch, (i_start, i_stop) in enumerate(chunk_indexes):
            out = func(*arg_tasks[ch], **karg_tasks[ch])
            output_list.append(out)

            if tmp_mode is None:
                for key, full_arr in assignment_dict.items():
                    out_chunk = out[key]
                    full_arr[i_start:i_stop] += out_chunk
            elif tmp_mode == 'memmap':
                pass
                # Nothing to do here because done inside the func with FuncThenAddChunk

    else:
        # parallel
        output_list = Parallel(n_jobs=n_jobs)(
            delayed(func)(*arg_task, **karg_task) for arg_task, karg_task in zip(arg_tasks, karg_tasks))

        if tmp_mode == 'memmap':
            pass
            # Nothing to do here because done inside the func
        else:
            # This case is very unefficient because it double the memory usage!!!!!!!
            for ch, (i_start, i_stop) in enumerate(chunk_indexes):
                for key, full_arr in assignment_dict.items():
                    full_arr[i_start:i_stop] += output_list[ch][key]

    return output_list



"""
Test implementation using cell models of the Blue Brain Project with LFPy.
The example assumes that cell models available from
https://bbpnmc.epfl.ch/nmc-portal/downloads are unzipped in the folder 'cell_models'

The function compile_all_mechanisms must be run once before any cell simulation
"""

import os
from os.path import join
import sys
from glob import glob
import numpy as np
import MEAutility as mu
import yaml
import time
from pathlib import Path
from distutils.version import StrictVersion

if StrictVersion(yaml.__version__) >= StrictVersion('5.0.0'):
    use_loader = True
else:
    use_loader = False


def get_templatename(f):
    """
    Assess from hoc file the templatename being specified within

    Arguments
    ---------
    f : file, mode 'r'

    Returns
    -------
    templatename : str

    """
    templatename = None
    f = open("template.hoc", 'r')
    for line in f.readlines():
        if 'begintemplate' in line.split():
            templatename = line.split()[-1]
            continue
    return templatename


def compile_all_mechanisms(cell_folder, verbose=False):
    """ Attempt to set up a folder with all unique mechanism *.mod files and 
        compile them all. assumes all cell models are in a folder 'cell_models'
    
    Parameters:
    -----------
    cell_folder : str
        Path to cell folder
    """

    if not os.path.isdir(join(cell_folder, 'mods')):
        os.mkdir(join(cell_folder, 'mods'))

    neurons = [join(cell_folder, f) \
               for f in os.listdir(join(cell_folder)) \
               if f != 'mods']
    if verbose:
        print(neurons)

    for nrn in neurons:
        for nmodl in glob(join(nrn, 'mechanisms', '*.mod')):
            while not os.path.isfile(join(cell_folder, 'mods', os.path.split(nmodl)[-1])):
                if sys.platform == 'win32':
                    _command = 'copy'
                else:
                    _command = 'cp'
                if verbose:
                    print('{} {} {}'.format(_command, nmodl, join(cell_folder, 'mods')))
                os.system('{} {} {}'.format(_command, nmodl, join(cell_folder, 'mods')))
    starting_dir = os.getcwd()
    os.chdir(join(cell_folder, 'mods'))
    os.system('nrnivmodl')
    os.chdir(starting_dir)


def return_bbp_cell(cell_folder, end_T, dt, start_T, verbose=False):
    """ Function to load cell models
    
    Parameters:
    -----------
    cell_folder : string
        Path to folder with the BBP cell model
    end_T : float
        Simulation length [ms]
    dt: float
        Time step of simulation [ms]
    start_T: float
        Simulation start time (recording starts at 0 ms)

    Returns:
    --------
    cell : object
        LFPy cell object
    """
    import mpi4py.MPI
    import LFPy
    import neuron
    neuron.h.load_file("stdrun.hoc")
    neuron.h.load_file("import3d.hoc")

    cwd = os.getcwd()
    os.chdir(cell_folder)
    if verbose:
        print("Simulating ", cell_folder)

    neuron.load_mechanisms('../mods')

    f = open("template.hoc", 'r')
    templatename = get_templatename(f)
    f.close()

    f = open("biophysics.hoc", 'r')
    biophysics = get_templatename(f)
    f.close()

    f = open("morphology.hoc", 'r')
    morphology = get_templatename(f)
    f.close()

    # get synapses template name
    f = open(join("synapses", "synapses.hoc"), 'r')
    synapses = get_templatename(f)
    f.close()

    neuron.h.load_file('constants.hoc')
    if not hasattr(neuron.h, morphology):
        neuron.h.load_file(1, "morphology.hoc")

    if not hasattr(neuron.h, biophysics):
        neuron.h.load_file(1, "biophysics.hoc")

    if not hasattr(neuron.h, synapses):
        # load synapses
        neuron.h.load_file(1, join('synapses', 'synapses.hoc'))

    if not hasattr(neuron.h, templatename):
        neuron.h.load_file(1, "template.hoc")

    morphologyfile = os.listdir('morphology')[0]  # glob('morphology\\*')[0]

    # Instantiate the cell(s) using LFPy
    cell = LFPy.TemplateCell(morphology=join('morphology', morphologyfile),
                             templatefile=join('template.hoc'),
                             templatename=templatename,
                             templateargs=0,
                             tstop=end_T,
                             tstart=start_T,
                             dt=dt,
                             v_init=-70,
                             pt3d=True,
                             delete_sections=True,
                             verbose=True)

    os.chdir(cwd)
    return cell


def return_bbp_cell_morphology(cell_name, cell_folder, pt3d=False):
    """ Function to load cell models

    Parameters:
    -----------
    cell_name : string
        Name of the cell type.
    cell_folder : string
        Folder containing cell models.
    pt3d : bool
        If True detailed 3d morphology is used

    Returns:
    --------
    cell : object
        LFPy cell object
    """
    import mpi4py.MPI
    import LFPy

    if not os.path.isdir(join(cell_folder, cell_name)):
        raise NotImplementedError('Cell model %s is not found in %s' \
                                  % (cell_name, cell_folder))

    morphologyfile = os.listdir(join(cell_folder, cell_name, 'morphology'))[0]
    morphology = join(cell_folder, cell_name, 'morphology', morphologyfile)

    cell = LFPy.Cell(morphology=morphology, pt3d=pt3d)
    return cell


def find_spike_idxs(v, thresh=-30, find_max=30):
    """ Find spike indices
    
    Parameters:
    -----------
    v: array_like
        Membrane potential
    thresh: float (optional, default = -30)
        Threshold for spike detections
    find_max: int
        Number of sample to find spike maximum after detection crossing

    Returns:
    --------
    spikes : array_like
        Indices of spike peaks in the positive direction, i.e. spikes
    """
    spikes_th = [idx for idx in range(len(v) - 1) if v[idx] < thresh < v[idx + 1]]
    spikes = []
    for sp in spikes_th:
        max_idx = np.argmax(v[sp:sp + find_max])
        spikes.append(sp + max_idx)
    return spikes


def set_input(weight, dt, T, cell, delay, stim_length):
    """ Set current input synapse in soma
    
    Parameters:
    -----------
    weight : float
        Strength of input current [nA]
    dt : float
        Time step of simulation [ms]
    T : float
        Total simulation time [ms]
    cell : object
        Cell object from LFPy
    delay : float
        Delay for input,i.e. when to start the input [ms]
    stim_length: float
        Duration of injected current [ms]
    
    Returns:
    --------
    noiseVec :  NEURON vector
        NEURON vector of input current
    cell : object
        LFPy cell object
    synapse : NEURON synapse
        NEURON synapse object
    """
    import neuron
    tot_ntsteps = int(round(T / dt + 1))

    I = np.ones(tot_ntsteps) * weight
    # I[stim_idxs] = weight
    noiseVec = neuron.h.Vector(I)
    syn = None
    for sec in cell.allseclist:
        if 'soma' in sec.name():
            # syn = neuron.h.ISyn(0.5, sec=sec) 
            syn = neuron.h.IClamp(0.5, sec=sec)
    syn.dur = stim_length
    syn.delay = delay  # cell.tstartms
    noiseVec.play(syn._ref_amp, dt)

    return noiseVec, cell, syn


def run_cell_model(cell_model_folder, verbose=False, sim_folder=None, save=True, return_vi=False,
                   custom_return_cell_function=None, **kwargs):
    """ Run simulation and adjust input strength to have a certain number of 
        spikes (target_spikes[0] < num_spikes <= target_spikes[1]
        where target_spikes=[10,30] by default)

    Parameters:
    -----------
    cell_model_folder : string
        Path to folder where cell model is saved.
    verbose : bool
        If True, output is verbose
    save : bool
        If True, currents and membrane potentials are saved in 'sim_folder'. If False the function returns the simulated
        cell, the soma potentials of the spikes, and the transmembrane currents of the spikes
    sim_folder : string
        Data directory for transmembrane currents and membrane potential of the neuron.
    custom_return_cell_function : function
        Python function to to return an LFPy cell from the cell_model_folder
    **kwargs : keyword arguments
        Kwargs must include: 'sim_time', 'dt', 'delay', 'weights', 'target_spikes', 'cut_out', 'seed'
    Returns:
    --------
    cell : object
        LFPy cell object (if save is False)
    v : np.array
        Array (N_spikes x t) with soma membrane potential (if save is False)
    i : np.array
        Array (N_spikes x N_compartments x t) with transmembrane currents (if save is False)

    """
    cell_name = Path(cell_model_folder).parts[-1]

    if custom_return_cell_function is None:
        return_function = return_bbp_cell
    else:
        return_function = custom_return_cell_function

    if save:
        assert sim_folder is not None, "Specify 'save_sim_folder' argument!"
        if not os.path.isdir(sim_folder):
            os.makedirs(sim_folder)

        imem_files = [f for f in os.listdir(sim_folder) if 'imem' in f]
        vmem_files = [f for f in os.listdir(sim_folder) if 'vmem' in f]

        if not (np.any([cell_name in ifile for ifile in imem_files]) and
                np.any([cell_name in vfile for vfile in vmem_files])):

            np.random.seed(kwargs['seed'])
            T = kwargs['sim_time'] * 1000
            dt = kwargs['dt']
            cell = return_function(cell_model_folder, end_T=T, dt=dt, start_T=0)

            delay = kwargs['delay']
            stim_length = T - delay
            weights = kwargs['weights']
            weight = weights[0]
            target_spikes = kwargs['target_spikes']
            cuts = kwargs['cut_out']
            cut_out = [int(cuts[0] / dt), int(cuts[1] / dt)]

            num_spikes = 0

            i = 0
            while not target_spikes[0] < num_spikes <= target_spikes[1]:
                noiseVec, cell, syn = set_input(weight, dt, T, cell, delay, stim_length)
                cell.simulate(rec_imem=True)

                t = cell.tvec
                v = cell.somav
                t = t
                v = v

                spikes = find_spike_idxs(v[cut_out[0]:-cut_out[1]])
                spikes = list(np.array(spikes) + cut_out[0])
                num_spikes = len(spikes)

                if verbose:
                    print("Input weight: ", weight, " - Num Spikes: ", num_spikes)
                if num_spikes >= target_spikes[1]:
                    weight *= weights[0]
                elif num_spikes <= target_spikes[0]:
                    weight *= weights[1]

                i += 1
                if i >= 10:
                    sys.exit()

            t = t[0:(cut_out[0] + cut_out[1])] - t[cut_out[0]]
            # discard first spike
            i_spikes = np.zeros((num_spikes - 1, cell.totnsegs, len(t)))
            v_spikes = np.zeros((num_spikes - 1, len(t)))

            for idx, spike_idx in enumerate(spikes[1:]):
                spike_idx = int(spike_idx)
                v_spike = v[spike_idx - cut_out[0]:spike_idx + cut_out[1]]
                i_spike = cell.imem[:, spike_idx - cut_out[0]:spike_idx + cut_out[1]]
                i_spikes[idx, :, :] = i_spike
                v_spikes[idx, :] = v_spike

            if not os.path.isdir(sim_folder):
                os.makedirs(sim_folder)
            np.save(join(sim_folder, 'imem_%d_%s.npy' % (num_spikes - 1, cell_name)), i_spikes)
            np.save(join(sim_folder, 'vmem_%d_%s.npy' % (num_spikes - 1, cell_name)), v_spikes)

        else:
            if verbose:
                print('\n\n\nCell has already be simulated. Using stored membrane currents\n\n\n')
    else:
        np.random.seed(kwargs['seed'])
        T = kwargs['sim_time'] * 1000
        dt = kwargs['dt']
        cell = return_function(cell_model_folder, end_T=T, dt=dt, start_T=0)

        delay = kwargs['delay']
        stim_length = T - delay
        weights = kwargs['weights']
        weight = weights[0]
        target_spikes = kwargs['target_spikes']
        cuts = kwargs['cut_out']
        cut_out = [int(cuts[0] / dt), int(cuts[1] / dt)]

        num_spikes = 0

        i = 0
        while not target_spikes[0] < num_spikes <= target_spikes[1]:
            noiseVec, cell, syn = set_input(weight, dt, T, cell, delay, stim_length)
            cell.simulate(rec_imem=True)

            t = cell.tvec
            v = cell.somav
            t = t
            v = v

            spikes = find_spike_idxs(v[cut_out[0]:-cut_out[1]])
            spikes = list(np.array(spikes) + cut_out[0])
            num_spikes = len(spikes)

            if verbose:
                print("Input weight: ", weight, " - Num Spikes: ", num_spikes)
            if num_spikes >= target_spikes[1]:
                weight *= weights[0]
            elif num_spikes <= target_spikes[0]:
                weight *= weights[1]

            i += 1
            if i >= 10:
                sys.exit()

        t = t[0:(cut_out[0] + cut_out[1])] - t[cut_out[0]]
        # discard first spike
        i_spikes = np.zeros((num_spikes - 1, cell.totnsegs, len(t)))
        v_spikes = np.zeros((num_spikes - 1, len(t)))

        for idx, spike_idx in enumerate(spikes[1:]):
            spike_idx = int(spike_idx)
            v_spike = v[spike_idx - cut_out[0]:spike_idx + cut_out[1]]
            i_spike = cell.imem[:, spike_idx - cut_out[0]:spike_idx + cut_out[1]]
            i_spikes[idx, :, :] = i_spike
            v_spikes[idx, :] = v_spike

        return cell, v_spikes, i_spikes


def calculate_extracellular_potential(cell, mea, ncontacts=10, position=None, rotation=None):
    '''
    Calculates extracellular signal in uV on MEA object.

    Parameters
    ----------
    cell : LFPy Cell
        The simulated cell
    mea : MEA, str, or dict
        Mea object from MEAutility, string with probe name, or dict with probe info

    Returns
    -------
    v_ext : np.array
        Extracellular potential computed on the electrodes (n_elec, n_timestamps)
    '''
    import LFPy

    if isinstance(mea, str):
        mea_obj = mu.return_mea(mea)
    elif isinstance(mea, dict):
        mea_obj = mu.return_mea(info=mea)
    elif isinstance(mea, mu.core.MEA):
        mea_obj = mea
    else:
        raise Exception("")

    if ncontacts > 1:
        electrodes = LFPy.RecExtElectrode(cell, probe=mea_obj, n=ncontacts)
    else:
        electrodes = LFPy.RecExtElectrode(cell, probe=mea_obj)

    if position is not None:
        assert len(position) == 3, "'position' should be a 3d array"
        cell.set_pos(position[0], position[1], position[2])

    if rotation is not None:
        assert len(rotation) == 3, "'rotation' should be a 3d array"
        cell.set_rotation(x=rotation[0], y=rotation[1], z=rotation[2])

    electrodes.calc_lfp()

    # Reverse rotation to bring cell back into initial rotation state
    if rotation is not None:
        rev_rot = [-r for r in rotation]
        cell.set_rotation(rev_rot[0], rev_rot[1], rev_rot[2], rotation_order='zyx')

    return 1000 * electrodes.LFP


def calc_extracellular(cell_model_folder, load_sim_folder, save_sim_folder=None, seed=0, verbose=False, position=None,
                       save=True, custom_return_cell_function=None, **kwargs):
    """
    Loads data from previous cell simulation, and use results to generate
    arbitrary number of spikes above a certain noise level.

    Parameters:
    -----------
    cell_model_folder : string
        Path to folder where cell model is saved.
    model_type : string
        Cell model type (e.g. 'bbp')
    load_sim_folder : string
        Path to folder from which NEURON simulation results (currents, membrane potential) are loaded
    save_sim_folder : string
        Path to folder where to save EAP data
    position : array
        3D position of the soma (optional, default is None and the cell is randomly located within specified limits)
    custom_return_cell_function : function
        Python function to to return an LFPy cell from the cell_model_folder
    save : bool
        If True eaps are saved in the 'save_sim_folder'. If False eaps, positions, and rotations are returned as arrays
    **kwargs: keyword arguments
        Template generation parameters (use mr.get_default_template_parameters() to retrieve the arguments)

    Returns:
    --------
        nothing, but saves the result
    """
    import LFPy
    cell_name = os.path.split(cell_model_folder)[-1]
    cell_save_name = cell_name
    np.random.seed(seed)

    T = kwargs['sim_time'] * 1000
    dt = kwargs['dt']
    rotation = kwargs['rot']
    nobs = kwargs['n']
    ncontacts = kwargs['ncontacts']
    overhang = kwargs['overhang']
    x_lim = kwargs['xlim']
    y_lim = kwargs['ylim']
    z_lim = kwargs['zlim']
    min_amp = kwargs['min_amp']
    MEAname = kwargs['probe']
    drifting = kwargs['drifting']
    if drifting:
        max_drift = kwargs['max_drift']
        min_drift = kwargs['min_drift']
        drift_steps = kwargs['drift_steps']
        drift_x_lim = kwargs['drift_xlim']
        drift_y_lim = kwargs['drift_ylim']
        drift_z_lim = kwargs['drift_zlim']

    if custom_return_cell_function is None:
        return_function = return_bbp_cell
        model_type = 'bbp'
    else:
        return_function = custom_return_cell_function
        model_type = 'custom'

    cuts = kwargs['cut_out']
    cut_out = [int(cuts[0] / dt), int(cuts[1] / dt)]

    cell = return_function(cell_model_folder, end_T=T, dt=dt, start_T=0)

    # Load data from previous cell simulation
    imem_file = [f for f in os.listdir(load_sim_folder) if cell_name in f and 'imem' in f][0]
    vmem_file = [f for f in os.listdir(load_sim_folder) if cell_name in f and 'vmem' in f][0]
    i_spikes = np.load(join(load_sim_folder, imem_file))
    v_spikes = np.load(join(load_sim_folder, vmem_file))
    cell.tvec = np.arange(i_spikes.shape[-1]) * dt

    save_spikes = []
    save_pos = []
    save_rot = []
    save_offs = []
    target_num_spikes = int(nobs)

    # load MEA info
    elinfo = mu.return_mea_info(electrode_name=MEAname)

    # Create save folder
    if save:
        assert save_sim_folder is not None, "Specify 'save_sim_folder' argument!"
        sim_folder = join(save_sim_folder, rotation)
        save_folder = join(sim_folder, 'tmp_%d_%s' % (target_num_spikes, MEAname))

        if not os.path.isdir(save_folder):
            os.makedirs(save_folder)

    if verbose:
        print('Cell ', cell_save_name, ' extracellular spikes to be simulated')

    mea = mu.return_mea(MEAname)
    if ncontacts > 1:
        electrodes = LFPy.RecExtElectrode(cell, probe=mea, n=ncontacts)
    else:
        electrodes = LFPy.RecExtElectrode(cell, probe=mea)
    pos = mea.positions
    elec_x = pos[:, 0]
    elec_y = pos[:, 1]
    elec_z = pos[:, 2]

    if x_lim is None:
        x_lim = [float(np.min(elec_x) - overhang),
                 float(np.max(elec_x) + overhang)]
    if y_lim is None:
        y_lim = [float(np.min(elec_y) - overhang),
                 float(np.max(elec_y) + overhang)]
    if z_lim is None:
        z_lim = [float(np.min(elec_z) - overhang),
                 float(np.max(elec_z) + overhang)]

    ignored = 0
    saved = 0
    i = 0

    while len(save_spikes) < target_num_spikes:
        if i > 1000 * target_num_spikes:
            if verbose:
                print("Gave up finding spikes above noise level for %s" % cell_name)
            break
        spike_idx = np.random.randint(0, i_spikes.shape[0])  # Each cell has several spikes to choose from
        cell.imem = i_spikes[spike_idx, :, :]
        cell.somav = v_spikes[spike_idx, :]

        if not drifting:
            espikes, pos, rot, offs = return_extracellular_spike(cell=cell, cell_name=cell_name, model_type=model_type,
                                                                 electrodes=electrodes,
                                                                 limits=[x_lim, y_lim, z_lim], rotation=rotation,
                                                                 pos=position)
            # Method of Images for semi-infinite planes
            if elinfo['type'] == 'mea':
                espikes = espikes * 2

            if check_espike(espikes, min_amp):
                espikes = center_espike(espikes, cut_out)
                save_spikes.append(espikes)
                save_pos.append(pos)
                save_rot.append(rot)
                save_offs.append(offs)
                plot_spike = False
                if verbose:
                    print('Cell: ' + cell_name + ' Progress: [' +
                          str(len(save_spikes)) + '/' + str(target_num_spikes) + ']')
                saved += 1
        else:
            espikes, pos, rot, offs = return_extracellular_spike(cell=cell, cell_name=cell_name, model_type=model_type,
                                                                 electrodes=electrodes,
                                                                 limits=[x_lim, y_lim, z_lim], rotation=rotation,
                                                                 pos=position)

            # Method of Images for semi-infinite planes
            if elinfo['type'] == 'mea':
                espikes = espikes * 2

            if x_lim[0] < pos[0] - drift_x_lim[0] < x_lim[1] and \
                    y_lim[0] < pos[1] - drift_y_lim[0] < y_lim[1] and \
                    z_lim[0] < pos[2] - drift_z_lim[0] < z_lim[1]:
                if check_espike(espikes, min_amp):
                    drift_ok = False
                    # fix rotation while drifting
                    cell.set_rotation(rot[0], rot[1], rot[2])
                    max_trials = 100
                    tr = 0
                    while not drift_ok and tr < max_trials:
                        init_pos = pos
                        # find drifting final position within drift limits
                        x_rand = np.random.uniform(init_pos[0] + drift_x_lim[0], init_pos[0] + drift_x_lim[1])
                        y_rand = np.random.uniform(init_pos[1] + drift_y_lim[0], init_pos[1] + drift_y_lim[1])
                        z_rand = np.random.uniform(init_pos[2] + drift_z_lim[0], init_pos[2] + drift_z_lim[1])
                        final_pos = [x_rand, y_rand, z_rand]
                        drift_dist = np.linalg.norm(np.array(init_pos) - np.array(final_pos))

                        # check location and boundaries
                        if max_drift > drift_dist > min_drift and \
                                x_lim[0] < x_rand < x_lim[1] and \
                                y_lim[0] < y_rand < y_lim[1] and \
                                z_lim[0] < z_rand < z_lim[1]:
                            espikes, pos, rot_, offs = return_extracellular_spike(cell=cell, cell_name=cell_name,
                                                                                  model_type=model_type,
                                                                                  electrodes=electrodes,
                                                                                  limits=[x_lim, y_lim, z_lim],
                                                                                  rotation=None,
                                                                                  pos=final_pos)
                            # check final position spike amplitude
                            if check_espike(espikes, min_amp):
                                if verbose:
                                    print('Found final drifting position')
                                drift_ok = True
                            else:
                                tr += 1
                                pass
                        else:
                            tr += 1
                            pass

                    # now compute drifting templates
                    if drift_ok:
                        drift_spikes = []
                        drift_pos = []
                        drift_rot = []
                        drift_dist = np.linalg.norm(np.array(init_pos) - np.array(final_pos))
                        drift_dir = np.array(final_pos) - np.array(init_pos)
                        for i, dp in enumerate(np.linspace(0, 1, drift_steps)):
                            pos_drift = init_pos + dp * drift_dir
                            espikes, pos, r_, offs = return_extracellular_spike(cell=cell, cell_name=cell_name,
                                                                                model_type=model_type,
                                                                                electrodes=electrodes,
                                                                                limits=[x_lim, y_lim, z_lim],
                                                                                rotation=None,
                                                                                pos=pos_drift)
                            espikes = center_espike(espikes, cut_out)
                            drift_spikes.append(espikes)
                            drift_pos.append(pos)
                            drift_rot.append(rot)

                        # reverse rotation
                        rev_rot = [-r for r in rot]
                        cell.set_rotation(rev_rot[0], rev_rot[1], rev_rot[2], rotation_order='zyx')

                        drift_spikes = np.array(drift_spikes)
                        drift_pos = np.array(drift_pos)
                        if verbose:
                            print('Drift done from ', init_pos, ' to ', final_pos, ' with ', drift_steps, ' steps')

                        save_spikes.append(drift_spikes)
                        save_pos.append(drift_pos)
                        save_rot.append(rot)
                        save_offs.append(offs)
                        if verbose:
                            print('Cell: ' + cell_name + ' Progress: [' + str(len(save_spikes)) + '/' +
                                  str(target_num_spikes) + ']')
                        saved += 1
                    else:
                        if verbose:
                            print('Discarded for trials')
                else:
                    pass
            else:
                if verbose:
                    print('Discarded position: ', pos)
        i += 1

    save_spikes = np.array(save_spikes)
    save_pos = np.array(save_pos)
    save_rot = np.array(save_rot)

    if save:
        np.save(join(save_folder, 'eap-%s' % cell_save_name), save_spikes)
        np.save(join(save_folder, 'pos-%s' % cell_save_name), save_pos)
        np.save(join(save_folder, 'rot-%s' % cell_save_name), save_rot)
    else:
        return save_spikes, save_pos, save_rot


def check_espike(espikes, min_amp):
    """
    Check extracellular spike amplitude and shape (neg peak > pos peak)

    Parameters
    ----------
    espike: np.array
        EAP (n_elec, n_samples)
    min_amp: float
        Minimum amplitude

    Returns
    -------
    valid: bool
        If True EAP is valid

    """
    valid = True
    if np.max(np.abs(np.min(espikes))) < min_amp:
        valid = False
    if np.abs(np.min(espikes)) < np.abs(np.max(espikes)):
        valid = False
    return valid


def center_espike(espike, cut_out_samples, tol=1):
    """
    Centers extracellular spike if the peak is not aligned.

    Parameters
    ----------
    espike : np.array
        EAP (n_elec, n_samples)
    cut_out_samples : list
        Samples before and after the peak
    tol : int
        Tolerance in number of samples

    Returns
    -------
    espike_centered: np.array
        Centered EAP (n_elec, n_samples)
    """
    expexted_peak = cut_out_samples[0]
    peak_idx = np.unravel_index(np.argmin(espike), espike.shape)[1]

    if np.abs(expexted_peak - peak_idx) >= tol:
        if expexted_peak - peak_idx < 0:
            diff = peak_idx - expexted_peak
            cent_espike = np.zeros_like(espike)
            cent_espike[:, :-diff] = espike[:, diff:]
            cent_espike[:, -diff:] = np.tile(espike[:, -1, np.newaxis], [1, diff])
        elif expexted_peak - peak_idx > 0:
            diff = expexted_peak - peak_idx
            cent_espike = np.zeros_like(espike)
            cent_espike[:, diff:] = espike[:, :-diff]
            cent_espike[:, :diff] = np.tile(espike[:, 0, np.newaxis], [1, diff])
    else:
        cent_espike = espike

    cent_peak_idx = np.unravel_index(np.argmin(cent_espike), cent_espike.shape)[1]
    assert np.abs(expexted_peak - cent_peak_idx) < tol, "Something went wrong in centering the spike"

    return cent_espike


def get_physrot_specs(cell_name, model):
    """  Return physrot specifications for cell types

    Parameters:
    -----------
    cell_name : string
        The name of the cell.
    Returns:
    --------
    polarlim : array_like
        lower and upper bound for the polar angle
    pref_orient : array_like
        3-dim vetor of preferred orientation
    """
    if model == 'bbp':
        polarlim = {'BP': [0., 15.],
                    'AC': None,
                    'BTC': None,  # [0.,15.],
                    'ChC': None,  # [0.,15.],
                    'DBC': None,  # [0.,15.],
                    'LBC': None,  # [0.,15.],
                    'MC': [0., 15.],
                    'NBC': None,
                    'NGC': None,
                    'SBC': None,
                    'PC': [0., 15.],
                    'SS': [0., 15.],
                    'SP': [0., 15.]}
        # how it's implemented, the NMC y axis points into the pref_orient direction after rotation
        pref_orient = {'BP': [0., 0., 1.],
                       'AC': None,
                       'BTC': None,  # [0.,0.,1.],
                       'ChC': None,  # [0.,0.,1.],
                       'DBC': None,  # [0.,0.,1.],
                       'LBC': None,  # [0.,0.,1.],
                       'MC': [0., 0., 1.],
                       'NBC': None,
                       'NGC': None,
                       'SBC': None,
                       'PC': [0., 0., 1.],
                       'SS': [0., 0., 1.],
                       'SP': [0., 0., 1.]}
        polar = None
        orient = None
        for k in polarlim.keys():
            if k in cell_name.split('_')[1]:
                polar = polarlim[k]
                orient = pref_orient[k]
                break
        return polar, orient
    else:
        raise NotImplementedError('Cell model %s is not implemented' % model)


def return_extracellular_spike(cell, cell_name, model_type,
                               electrodes, limits, rotation, pos=None):
    """    Calculate extracellular spike on MEA 
           at random position relative to cell

    Parameters:
    -----------
    cell: object
        cell object from LFPy
    cell_name: string
        name of cell model
    electrodes: LFPyRecExtElectrode
        The LFPy electrode object
    limits: array_like
        boundaries for neuron locations, shape=(3,2)
    rotation: string 
        random rotation to apply to the neuron ('Norot', '3drot', 'physrot')
    pos: array_like, (optional, default None)
        Can be used to set the cell soma to a specific position. If ``None``,
        the random position is used.
    Returns:
    --------
    Extracellular spike for each MEA contact site
    """
    import LFPy

    def get_xyz_angles(R):
        """ Get rotation angles for each axis from rotation matrix
        
        Parameters;
        -----------
        R : matrix
            3x3 rotation matrix

        Returns:
        --------
        R_z : float
        R_y : float
        R_x : float
            Three angles for rotations around axis, defined by R = R_z.R_y.R_x
        """
        rot_x = np.arctan2(R[2, 1], R[2, 2])
        rot_y = np.arcsin(-R[2, 0])
        rot_z = np.arctan2(R[1, 0], R[0, 0])
        return rot_x, rot_y, rot_z

    def get_rnd_rot_Arvo():
        """ Generate uniformly distributed random rotation matrices
        see: 'Fast Random Rotation Matrices' by Arvo (1992)
        
        Returns:
        --------
        R : 3x3 matrix
            random rotation matrix
        """
        gamma = np.random.uniform(0, 2. * np.pi)
        rotation_z = np.matrix([[np.cos(gamma), -np.sin(gamma), 0],
                                [np.sin(gamma), np.cos(gamma), 0],
                                [0, 0, 1]])
        x = np.random.uniform(size=2)
        v = np.array([np.cos(2. * np.pi * x[0]) * np.sqrt(x[1]),
                      np.sin(2. * np.pi * x[0]) * np.sqrt(x[1]),
                      np.sqrt(1 - x[1])])
        H = np.identity(3) - 2. * np.outer(v, v)
        M = -np.dot(H, rotation_z)
        return M

    def check_solidangle(matrix, pre, post, polarlim):
        """ Check whether a matrix rotates the vector 'pre' into a region
            defined by 'polarlim' around the vector 'post'

        Parameters:
        -----------
        matrix : matrix
            3x3 rotation matrix
        pre : array_like
            3-dim vector to be rotated
        post : array_like
            axis of the cones defining the post-rotation region
        polarlim : [float,float]
            Angles specifying the opening of the inner and outer cone 
            (aperture = 2*polarlim),
            i.e. the angle between rotated pre vector and post vector has to ly 
            within these polar limits.  

        Returns:
        --------
        test : bool
            True if the vector np.dot(matrix,pre) lies inside the specified region.
        """
        postest = np.dot(matrix, pre)
        c = np.dot(post / np.linalg.norm(post), postest / np.linalg.norm(postest))
        if np.cos(np.deg2rad(polarlim[1])) <= c <= np.cos(np.deg2rad(polarlim[0])):
            return True
        else:
            return False

    # rotate neuron
    if rotation == 'norot':
        if model_type == 'bbp':
            # orientate cells in z direction
            x_rot_offset = np.pi / 2.
            y_rot_offset = 0
            z_rot_offset = 0
        else:
            x_rot_offset = 0
            y_rot_offset = 0
            z_rot_offset = 0
        x_rot = x_rot_offset
        y_rot = y_rot_offset
        z_rot = z_rot_offset
    elif rotation == 'xrot':
        if model_type == 'bbp':
            # orientate cells in z direction
            x_rot_offset = np.pi / 2.
            y_rot_offset = 0
            z_rot_offset = 0
        else:
            x_rot_offset = 0
            y_rot_offset = 0
            z_rot_offset = 0
        x_rot, _, _ = get_xyz_angles(np.array(get_rnd_rot_Arvo()))
        x_rot = x_rot + x_rot_offset
        y_rot = y_rot_offset
        z_rot = z_rot_offset
    elif rotation == 'yrot':
        if model_type == 'bbp':
            # orientate cells in z direction
            x_rot_offset = np.pi / 2.
            y_rot_offset = 0
            z_rot_offset = 0
        else:
            x_rot_offset = 0
            y_rot_offset = 0
            z_rot_offset = 0
        _, y_rot, _ = get_xyz_angles(np.array(get_rnd_rot_Arvo()))
        x_rot = x_rot_offset
        y_rot = y_rot + y_rot_offset
        z_rot = z_rot_offset
    elif rotation == 'zrot':
        if model_type == 'bbp':
            # orientate cells in z direction
            x_rot_offset = np.pi / 2.
            y_rot_offset = 0
            z_rot_offset = 0
        else:
            x_rot_offset = 0
            y_rot_offset = 0
            z_rot_offset = 0
        _, _, z_rot = get_xyz_angles(np.array(get_rnd_rot_Arvo()))
        x_rot = x_rot_offset
        y_rot = y_rot_offset
        z_rot = z_rot + z_rot_offset
    elif rotation == '3drot':
        if model_type == 'bbp':
            x_rot_offset = np.pi / 2.  # align neuron with z axis
            y_rot_offset = 0  # align neuron with z axis
            z_rot_offset = 0  # align neuron with z axis
        else:
            x_rot_offset = 0
            y_rot_offset = 0
            z_rot_offset = 0
        x_rot, y_rot, z_rot = get_xyz_angles(np.array(get_rnd_rot_Arvo()))
        x_rot = x_rot + x_rot_offset
        y_rot = y_rot + y_rot_offset
        z_rot = z_rot + z_rot_offset
    elif rotation == 'physrot':
        polarlim, pref_orient = get_physrot_specs(cell_name, model_type)
        if model_type == 'bbp':
            x_rot_offset = np.pi / 2.  # align neuron with z axis
            y_rot_offset = 0  # align neuron with z axis
            z_rot_offset = 0  # align neuron with z axis
        else:
            raise NotImplementedError("'physrot' rotation is only available with BBP cells")
        while True:
            R = np.array(get_rnd_rot_Arvo())
            if polarlim is None or pref_orient is None:
                valid = True
            else:
                valid = check_solidangle(R, [0., 0., 1.], pref_orient, polarlim)
            if valid:
                x_rot, y_rot, z_rot = get_xyz_angles(R)
                x_rot = x_rot + x_rot_offset
                y_rot = y_rot + y_rot_offset
                z_rot = z_rot + z_rot_offset
                break
    else:
        rotation = None
        x_rot = 0
        y_rot = 0
        z_rot = 0

    if pos is None:
        """Move neuron randomly"""
        x_rand = np.random.uniform(limits[0][0], limits[0][1])
        y_rand = np.random.uniform(limits[1][0], limits[1][1])
        z_rand = np.random.uniform(limits[2][0], limits[2][1])
        cell.set_pos(x_rand, y_rand, z_rand)
        pos = [x_rand, y_rand, z_rand]
    else:
        cell.set_pos(pos[0], pos[1], pos[2])
    cell.set_rotation(x=x_rot, y=y_rot, z=z_rot)
    rot = [x_rot, y_rot, z_rot]

    electrodes.calc_lfp()

    # Reverse rotation to bring cell back into initial rotation state
    if rotation is not None:
        rev_rot = [-r for r in rot]
        cell.set_rotation(rev_rot[0], rev_rot[1], rev_rot[2], rotation_order='zyx')

    return 1000 * electrodes.LFP, pos, rot, electrodes.offsets


def str2bool(v):
    """ Transform string to bool
    
    Parameters:
    -----------
    v : str
    
    Returns:
    --------
    transformed_v, bool
        If v is any of ("yes", "true", "t", "1") (case insensitive) 
        ``True`` is returned, else ``False``
    """
    return v.lower() in ("yes", "true", "t", "1")


def simulate_templates_one_cell(cell_model, intra_save_folder, params, verbose, custom_return_cell_function=None):
    '''

    Parameters
    ----------
    cell_model
    intra_save_folder
    params
    verbose
    custom_return_cell_function

    Returns
    -------

    '''
    run_cell_model(cell_model, save=True, sim_folder=intra_save_folder, verbose=verbose,
                   custom_return_cell_function=custom_return_cell_function, **params)
    print('Extracellular simulation: ', cell_model)
    eaps, locs, rots = calc_extracellular(cell_model, intra_save_folder, verbose=verbose,
                                          save=False, custom_return_cell_function=custom_return_cell_function, **params)

    return eaps, locs, rots


if __name__ == '__main__':

    if len(sys.argv) == 3 and sys.argv[1] == 'compile':
        cell_folder = sys.argv[2]
        compile_all_mechanisms(cell_folder)
        sys.exit(0)
    elif len(sys.argv) == 5:
        cell_model = sys.argv[1]
        intraonly = str2bool(sys.argv[2])
        params_path = sys.argv[3]
        verbose = str2bool(sys.argv[4])

        with open(params_path, 'r') as f:
            if use_loader:
                params = yaml.load(f, Loader=yaml.FullLoader)
            else:
                params = yaml.load(f)

        sim_folder = params['templates_folder']
        cell_folder = params['cell_models_folder']
        rot = params['rot']

        extra_sim_folder = params['templates_folder']
        vm_im_sim_folder = join(params['templates_folder'], 'intracellular')

        print('Intracellular simulation: ', cell_model)
        run_cell_model(cell_model_folder=cell_model, sim_folder=vm_im_sim_folder, verbose=verbose, **params)
        if not intraonly:
            print('Extracellular simulation: ', cell_model)
            calc_extracellular(cell_model_folder=cell_model, save_sim_folder=extra_sim_folder,
                               load_sim_folder=vm_im_sim_folder, verbose=verbose, **params)



import pytest
import sys
import os
import numpy as np
import unittest
import MEArec as mr
import tempfile
import shutil
import yaml
import elephant.statistics as stat
from distutils.version import StrictVersion
import tempfile
from click.testing import CliRunner

from MEArec.cli import cli

if StrictVersion(yaml.__version__) >= StrictVersion('5.0.0'):
    use_loader = True
else:
    use_loader = False

local_temp = False


class TestGenerators(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        info, info_folder = mr.get_default_config()
        cell_models_folder = mr.get_default_cell_models_folder()
        self.num_cells = len([f for f in os.listdir(cell_models_folder) if 'mods' not in f])
        self.n = 10
        self.n_drift = 5

        templates_params = mr.get_default_templates_params()
        rec_params = mr.get_default_recordings_params()

        if not local_temp:
            self.test_dir = tempfile.mkdtemp()
        else:
            self.test_dir = './tmp'

        if not (os.path.isfile(self.test_dir + '/templates.h5')
                and os.path.isfile(self.test_dir + '/templates_drift.h5')):
            templates_params['n'] = self.n
            templates_params['ncontacts'] = 1
            templates_params['probe'] = 'Neuronexus-32'
            templates_folder = info['templates_folder']
            print('Generating non-drifting templates')
            templates_params['min_amp'] = 10
            print(templates_params)
            self.tempgen = mr.gen_templates(cell_models_folder, templates_tmp_folder=templates_folder,
                                            params=templates_params, parallel=True, delete_tmp=True, verbose=True)
            self.templates_params = templates_params
            self.num_templates, self.num_chan, self.num_samples = self.tempgen.templates.shape

            templates_params['drifting'] = True
            templates_params['n'] = self.n_drift
            templates_params['drift_steps'] = 10
            templates_params['rot'] = 'norot'
            templates_params['min_amp'] = 10
            print('Generating drifting templates')
            self.tempgen_drift = mr.gen_templates(cell_models_folder, templates_tmp_folder=templates_folder,
                                                  params=templates_params, parallel=True, delete_tmp=True, verbose=True)
            self.templates_params_drift = templates_params
            self.num_steps_drift = self.tempgen_drift.templates.shape[1]

            print('Making test recordings to test load functions')
            mr.save_template_generator(self.tempgen, self.test_dir + '/templates.h5')
            mr.save_template_generator(self.tempgen_drift, self.test_dir + '/templates_drift.h5')
        else:
            print(self.test_dir + '/templates.h5')
            self.tempgen = mr.load_templates(self.test_dir + '/templates.h5', return_h5_objects=False)
            self.tempgen_drift = mr.load_templates(self.test_dir + '/templates_drift.h5')
            self.templates_params = self.tempgen.info['params']
            self.templates_params_drift = self.tempgen_drift.info['params']
            self.num_steps_drift = self.tempgen_drift.templates.shape[1]
            self.num_templates, self.num_chan, self.num_samples = self.tempgen.templates.shape

        if not os.path.isfile(self.test_dir + '/recordings.h5'):
            ne = 2
            ni = 1
            rec_params['spiketrains']['n_exc'] = ne
            rec_params['spiketrains']['n_inh'] = ni
            rec_params['recordings']['modulation'] = 'none'
            rec_params['recordings']['filter'] = False
            rec_params['recordings']['extract_waveforms'] = True
            rec_params['recordings']['overlap'] = True
            rec_params['spiketrains']['duration'] = 2
            rec_params['templates']['min_dist'] = 1

            self.recgen = mr.gen_recordings(params=rec_params, tempgen=self.tempgen, verbose=False)
            self.recgen.annotate_overlapping_spikes(parallel=False)
            mr.save_recording_generator(self.recgen, self.test_dir + '/recordings.h5')
        else:
            self.recgen = mr.load_recordings(self.test_dir + '/recordings.h5', return_h5_objects=False)

    @classmethod
    def tearDownClass(self):
        # Remove the directory after the test
        if not local_temp:
            shutil.rmtree(self.test_dir)

    def test_gen_templates(self):
        print('Test templates generation')
        n = self.n
        num_cells = self.num_cells
        templates_params = self.templates_params

        assert self.tempgen.templates.shape[0] == (n * num_cells)
        assert len(self.tempgen.locations) == (n * num_cells)
        assert len(self.tempgen.rotations) == (n * num_cells)
        assert len(self.tempgen.celltypes) == (n * num_cells)
        assert len(np.unique(self.tempgen.celltypes)) == num_cells
        assert np.min(self.tempgen.locations[:, 0]) > templates_params['xlim'][0] \
               and np.max(self.tempgen.locations[:, 0]) < templates_params['xlim'][1]

    def test_gen_templates_drift(self):
        print('Test drifting templates generation')
        n = self.n_drift
        num_cells = self.num_cells
        n_steps = self.num_steps_drift
        templates_params = self.templates_params_drift

        assert self.tempgen_drift.templates.shape[0] == (n * num_cells)
        assert self.tempgen_drift.locations.shape == (n * num_cells, n_steps, 3)
        assert len(self.tempgen_drift.rotations) == (n * num_cells)
        assert len(self.tempgen_drift.celltypes) == (n * num_cells)
        assert len(np.unique(self.tempgen_drift.celltypes)) == num_cells
        assert np.min(self.tempgen_drift.locations[:, :, 0]) > templates_params['xlim'][0] \
               and np.max(self.tempgen_drift.locations[:, :, 0]) < templates_params['xlim'][1]
        assert self.tempgen_drift.templates.shape[1] == self.num_steps_drift

    def test_gen_spiketrains(self):
        print('Test spike train generation')
        rec_params = mr.get_default_recordings_params()
        sp_params = rec_params['spiketrains']
        spgen = mr.SpikeTrainGenerator(sp_params, seed=0)
        spgen.generate_spikes()

        # check ref period
        for st in spgen.spiketrains:
            isi = stat.isi(st).rescale('ms')
            assert np.all(isi.magnitude > sp_params['ref_per'])
            assert (1 / np.mean(isi.rescale('s'))) > sp_params['min_rate']

        sp_params['process'] = 'gamma'
        spgen = mr.SpikeTrainGenerator(sp_params, seed=0)
        spgen.generate_spikes()
        for st in spgen.spiketrains:
            isi = stat.isi(st).rescale('ms')
            assert np.all(isi.magnitude > sp_params['ref_per'])
            assert (1 / np.mean(isi.rescale('s'))) > sp_params['min_rate']

        spgen = mr.gen_spiketrains(sp_params, seed=0)
        spiketrains = spgen.spiketrains
        spgen_st = mr.gen_spiketrains(spiketrains=spiketrains)
        for (st, st_) in zip(spgen.spiketrains, spgen_st.spiketrains):
            assert np.allclose(st.times.magnitude, st_.times.magnitude)

    def test_gen_recordings_with_spiketrains(self):
        print('Test recording generation - from spiketrains')
        rec_params = mr.get_default_recordings_params()
        sp_params = rec_params['spiketrains']
        sp_params['n_exc'] = 2
        sp_params['n_inh'] = 1
        sp_params['duration'] = 5
        spgen = mr.gen_spiketrains(sp_params)
        recgen = mr.gen_recordings(params=rec_params, spgen=spgen, tempgen=self.tempgen, verbose=False)

        for (st, st_) in zip(spgen.spiketrains, recgen.spiketrains):
            assert np.allclose(st.times.magnitude, st_.times.magnitude)
        del recgen

    def test_gen_recordings_modulations(self):
        print('Test recording generation - modulation')
        ne = 1
        ni = 2
        fe = 30
        fi = 50
        num_chan = self.num_chan
        n_neurons = ne + ni

        rec_params = mr.get_default_recordings_params()

        rec_params['spiketrains']['n_exc'] = ne
        rec_params['spiketrains']['n_inh'] = ni
        rec_params['spiketrains']['f_exc'] = fe
        rec_params['spiketrains']['f_inh'] = fi
        rec_params['spiketrains']['duration'] = 3
        n_jitter = [1, 5]
        modulations = ['none', 'electrode', 'template']
        chunk_rec = [0, 2]

        rec_params['templates']['min_dist'] = 1
        bursting = [False, True]

        for mod in modulations:
            for b in bursting:
                for j in n_jitter:
                    for ch in chunk_rec:
                        print('Modulation: modulation', mod, 'bursting', b, 'jitter', j, 'chunk', ch)
                        rec_params['templates']['n_jitters'] = j
                        rec_params['recordings']['modulation'] = mod
                        rec_params['recordings']['bursting'] = b
                        rec_params['recordings']['chunk_duration'] = ch

                        if mod == 'electrode' and b is True and j == 5:
                            rec_params['cell_types'] = None

                        recgen_mod = mr.gen_recordings(params=rec_params, tempgen=self.tempgen, verbose=False)

                        assert recgen_mod.recordings.shape[0] == num_chan
                        assert len(recgen_mod.spiketrains) == n_neurons
                        assert recgen_mod.channel_positions.shape == (num_chan, 3)
                        if j == 1:
                            assert recgen_mod.templates.shape[:2] == (n_neurons, num_chan)
                        else:
                            assert recgen_mod.templates.shape[:3] == (n_neurons, j, num_chan)
                        assert recgen_mod.voltage_peaks.shape == (n_neurons, num_chan)
                        assert len(recgen_mod.spike_traces) == n_neurons
                        del recgen_mod

    def test_gen_recordings_bursting(self):
        print('Test recording generation - shape_mod')
        rates = [30, 50]
        types = ['E', 'I']
        num_chan = self.num_chan
        n_neurons = len(rates)

        rec_params = mr.get_default_recordings_params()

        rec_params['spiketrains']['rates'] = rates
        rec_params['spiketrains']['types'] = types
        rec_params['spiketrains']['duration'] = 3
        n_jitter = 4
        rec_params['templates']['n_jitters'] = n_jitter
        rec_params['recordings']['modulation'] = 'electrode'
        rec_params['recordings']['bursting'] = True
        rec_params['recordings']['shape_mod'] = True
        rec_params['recordings']['sync_rate'] = 0.2
        rec_params['recordings']['overlap'] = True
        rec_params['recordings']['extract_waveforms'] = True
        rec_params['templates']['min_dist'] = 1
        rec_params['templates']['min_amp'] = 0

        recgen_burst = mr.gen_recordings(params=rec_params, tempgen=self.tempgen, verbose=False)
        recgen_burst.extract_waveforms()

        assert recgen_burst.recordings.shape[0] == num_chan
        assert len(recgen_burst.spiketrains) == n_neurons
        assert recgen_burst.channel_positions.shape == (num_chan, 3)
        assert recgen_burst.templates.shape[:3] == (n_neurons, n_jitter, num_chan)
        assert recgen_burst.voltage_peaks.shape == (n_neurons, num_chan)
        assert len(recgen_burst.spike_traces) == n_neurons
        del recgen_burst

        rec_params['recordings']['modulation'] = 'template'
        rec_params['recordings']['n_bursting'] = 2

        recgen_burst = mr.gen_recordings(params=rec_params, tempgen=self.tempgen, verbose=False)
        recgen_burst.extract_waveforms()

        assert recgen_burst.recordings.shape[0] == num_chan
        assert len(recgen_burst.spiketrains) == n_neurons
        assert recgen_burst.channel_positions.shape == (num_chan, 3)
        assert recgen_burst.templates.shape[:3] == (n_neurons, n_jitter, num_chan)
        assert recgen_burst.voltage_peaks.shape == (n_neurons, num_chan)
        assert len(recgen_burst.spike_traces) == n_neurons
        del recgen_burst

    def test_gen_recordings_noise(self):
        print('Test recording generation - noise')
        ne = 0
        ni = 0
        num_chan = self.num_chan
        n_neurons = ne + ni

        rec_params = mr.get_default_recordings_params()

        rec_params['spiketrains']['n_exc'] = ne
        rec_params['spiketrains']['n_inh'] = ni
        rec_params['spiketrains']['duration'] = 3
        n_jitter = 3
        rec_params['templates']['n_jitters'] = n_jitter
        rec_params['templates']['min_dist'] = 1
        rec_params['recordings']['modulation'] = 'none'
        noise_modes = ['uncorrelated', 'distance-correlated']
        chunk_noise = [0, 2]
        noise_color = [True, False]
        noise_level = 10
        rec_params['recordings']['noise_level'] = noise_level
        rec_params['recordings']['filter'] = False

        for mode in noise_modes:
            for ch in chunk_noise:
                for color in noise_color:
                    print('Noise: mode', mode, 'chunks', ch, 'color', color)
                    rec_params['recordings']['noise_mode'] = mode
                    rec_params['recordings']['chunk_duration'] = ch
                    rec_params['recordings']['noise_color'] = color
                    recgen_noise = mr.gen_recordings(params=rec_params, tempgen=self.tempgen, verbose=False)

                    if mode == 'uncorrelated' and ch == 0 and not noise_color:
                        rec_params['recordings']['fs'] = 30000
                        rec_params['recordings']['chunk_duration'] = 1
                    if noise_color:
                        rec_params['recordings']['filter_cutoff'] = 500

                    assert recgen_noise.recordings.shape[0] == num_chan
                    assert len(recgen_noise.spiketrains) == n_neurons
                    assert recgen_noise.channel_positions.shape == (num_chan, 3)
                    assert len(recgen_noise.spike_traces) == n_neurons
                    assert np.isclose(np.std(recgen_noise.recordings), noise_level, atol=1)
                    del recgen_noise

    def test_gen_recordings_far_neurons(self):
        print('Test recording generation - far neurons')
        ne = 0
        ni = 0
        num_chan = self.num_chan
        n_neurons = ne + ni

        rec_params = mr.get_default_recordings_params()

        rec_params['spiketrains']['n_exc'] = ne
        rec_params['spiketrains']['n_inh'] = ni
        rec_params['spiketrains']['duration'] = 3
        n_jitter = 3
        rec_params['templates']['n_jitters'] = n_jitter
        rec_params['templates']['min_dist'] = 1
        rec_params['recordings']['modulation'] = 'none'
        rec_params['recordings']['noise_mode'] = 'far-neurons'
        rec_params['recordings']['far_neurons_n'] = 10
        rec_params['recordings']['far_neurons_max_amp'] = 100
        rec_params['recordings']['far_neurons_exc_inh_ratio'] = 0.8
        noise_level = 20
        rec_params['recordings']['noise_level'] = noise_level
        rec_params['recordings']['filter'] = False
        recgen_noise = mr.gen_recordings(params=rec_params, tempgen=self.tempgen, verbose=False)

        assert recgen_noise.recordings.shape[0] == num_chan
        assert recgen_noise.channel_positions.shape == (num_chan, 3)
        assert len(recgen_noise.spiketrains) == n_neurons
        assert len(recgen_noise.spiketrains) == n_neurons
        assert len(recgen_noise.spiketrains) == n_neurons
        assert len(recgen_noise.spike_traces) == n_neurons
        assert np.isclose(np.std(recgen_noise.recordings), noise_level, atol=1)
        del recgen_noise

        rec_params['recordings']['chunk_duration'] = 1
        recgen_noise = mr.gen_recordings(params=rec_params, tempgen=self.tempgen, verbose=False)

        assert recgen_noise.recordings.shape[0] == num_chan
        assert recgen_noise.channel_positions.shape == (num_chan, 3)
        assert len(recgen_noise.spiketrains) == n_neurons
        assert len(recgen_noise.spiketrains) == n_neurons
        assert len(recgen_noise.spiketrains) == n_neurons
        assert len(recgen_noise.spike_traces) == n_neurons
        assert np.isclose(np.std(recgen_noise.recordings), noise_level, atol=1)
        del recgen_noise

    def test_gen_recordings_drift(self):
        print('Test recording generation - drift')
        ne = 1
        ni = 1
        num_chan = self.num_chan
        n_neurons = ne + ni

        rec_params = mr.get_default_recordings_params()

        rec_params['spiketrains']['n_exc'] = ne
        rec_params['spiketrains']['n_inh'] = ni
        rec_params['spiketrains']['duration'] = 5
        n_jitter = [1, 3]
        rec_params['recordings']['drifting'] = True
        rec_params['recordings']['drift_velocity'] = 500
        rec_params['templates']['min_dist'] = 1
        chunk_rec = [0, 2]

        modulations = ['none', 'template', 'electrode']
        bursting = [False, True]
        drift_modes = ['slow', 'fast', 'slow+fast']

        for i, mod in enumerate(modulations):
            for dm in drift_modes:
                for b in bursting:
                    for j in n_jitter:
                        for ch in chunk_rec:
                            print('Drifting: modulation', mod, 'bursting', b, 'jitter', j, 'drift mode', dm,
                                  'chunk', ch)
                            rec_params['templates']['n_jitters'] = j
                            rec_params['recordings']['modulation'] = mod
                            rec_params['recordings']['bursting'] = b
                            rec_params['recordings']['chunk_duration'] = ch
                            rec_params['recordings']['drift_mode'] = dm

                            if i == len(modulations) - 1:
                                rec_params['recordings']['fs'] = 30000
                                rec_params['recordings']['n_drifting'] = 1
                            if mod == 'electrode' and b is True and j == 5:
                                rec_params['cell_types'] = None
                                rec_params['recordings']['shape_mod'] = True
                            recgen_drift = mr.gen_recordings(params=rec_params, tempgen=self.tempgen_drift,
                                                             verbose=False)
                            assert recgen_drift.recordings.shape[0] == num_chan
                            assert len(recgen_drift.spiketrains) == n_neurons
                            assert recgen_drift.channel_positions.shape == (num_chan, 3)
                            if j == 1:
                                assert recgen_drift.templates.shape[0] == n_neurons
                                assert recgen_drift.templates.shape[2] == num_chan
                            else:
                                assert recgen_drift.templates.shape[0] == n_neurons
                                assert recgen_drift.templates.shape[2] == j
                                assert recgen_drift.templates.shape[3] == num_chan
                            assert len(recgen_drift.spike_traces) == n_neurons
                            del recgen_drift

    def test_save_load_templates(self):
        tempgen = mr.load_templates(self.test_dir + '/templates.h5', verbose=True)
        tempgen_drift = mr.load_templates(self.test_dir + '/templates_drift.h5')
        tempgen_drift_f = mr.load_templates(self.test_dir + '/templates_drift.h5', return_h5_objects=True)

        assert np.allclose(tempgen.templates, self.tempgen.templates)
        assert np.allclose(tempgen.locations, self.tempgen.locations)
        assert np.allclose(tempgen.rotations, self.tempgen.rotations)
        assert np.allclose(tempgen_drift.templates, self.tempgen_drift.templates)
        assert np.allclose(tempgen_drift.locations, self.tempgen_drift.locations)
        assert np.allclose(tempgen_drift.rotations, self.tempgen_drift.rotations)
        assert np.allclose(tempgen_drift.templates, np.array(tempgen_drift_f.templates))

    def test_save_load_recordings(self):
        recgen_loaded = mr.load_recordings(self.test_dir + '/recordings.h5', return_h5_objects=False, verbose=True, )
        recgen_loaded_f = mr.load_recordings(self.test_dir + '/recordings.h5', return_h5_objects=False, verbose=True)
        recgen_loaded_r = mr.load_recordings(self.test_dir + '/recordings.h5', load=['recordings'],
                                             return_h5_objects=True, verbose=True)

        assert np.allclose(recgen_loaded.templates, self.recgen.templates)
        assert np.allclose(recgen_loaded.recordings, self.recgen.recordings)
        assert np.allclose(recgen_loaded.spike_traces, self.recgen.spike_traces)
        assert np.allclose(recgen_loaded.voltage_peaks, self.recgen.voltage_peaks)
        assert np.allclose(recgen_loaded.channel_positions, self.recgen.channel_positions)
        assert np.allclose(recgen_loaded.timestamps.magnitude, self.recgen.timestamps.magnitude)
        assert np.allclose(recgen_loaded.recordings, np.array(recgen_loaded_f.recordings))
        assert np.allclose(recgen_loaded.recordings, np.array(recgen_loaded_r.recordings))

    def test_plots(self):
        _ = mr.plot_rasters(self.recgen.spiketrains)
        _ = mr.plot_rasters(self.recgen.spiketrains, overlap=True)
        _ = mr.plot_rasters(self.recgen.spiketrains, cell_type=True)
        _ = mr.plot_rasters(self.recgen.spiketrains, color='g')
        _ = mr.plot_rasters(self.recgen.spiketrains, color=['g'] * len(self.recgen.spiketrains))
        _ = mr.plot_recordings(self.recgen)
        _ = mr.plot_recordings(self.recgen, overlay_templates=True)
        _ = mr.plot_recordings(self.recgen, overlay_templates=True, max_channels_per_template=3)
        _ = mr.plot_templates(self.recgen)
        _ = mr.plot_templates(self.recgen, single_axes=True)
        _ = mr.plot_waveforms(self.recgen)
        _ = mr.plot_waveforms(self.recgen, electrode=0)
        _ = mr.plot_waveforms(self.recgen, electrode='max')
        _ = mr.plot_waveforms(self.recgen, electrode='max', max_waveforms=2)
        self.recgen.spiketrains[0].waveforms = None
        _ = mr.plot_waveforms(self.recgen, electrode='max')

    def test_extract_features(self):
        feat_t0 = mr.get_templates_features(self.tempgen.templates, feat_list=['neg', 'pos', 'amp', 'width', 'fwhm',
                                                                               'ratio', 'speed'],
                                            dt=self.tempgen.info['params']['dt'])
        assert feat_t0['neg'].shape == (self.tempgen.templates.shape[0], self.tempgen.templates.shape[1])
        assert feat_t0['pos'].shape == (self.tempgen.templates.shape[0], self.tempgen.templates.shape[1])
        assert feat_t0['amp'].shape == (self.tempgen.templates.shape[0], self.tempgen.templates.shape[1])
        assert feat_t0['width'].shape == (self.tempgen.templates.shape[0], self.tempgen.templates.shape[1])
        assert feat_t0['fwhm'].shape == (self.tempgen.templates.shape[0], self.tempgen.templates.shape[1])
        assert feat_t0['ratio'].shape == (self.tempgen.templates.shape[0], self.tempgen.templates.shape[1])
        assert feat_t0['speed'].shape == (self.tempgen.templates.shape[0], self.tempgen.templates.shape[1])

    def test_recordings_resample(self):
        print('Test recording generation - resampling')
        ne = 2
        ni = 1
        num_chan = self.num_chan
        n_neurons = ne + ni
        fs = 10000
        duration = 5

        rec_params = mr.get_default_recordings_params()

        rec_params['spiketrains']['n_exc'] = ne
        rec_params['spiketrains']['n_inh'] = ni
        rec_params['spiketrains']['duration'] = duration
        n_jitter = 3
        rec_params['templates']['n_jitters'] = n_jitter
        rec_params['recordings']['modulation'] = 'none'
        rec_params['recordings']['fs'] = fs
        recgen_rs = mr.gen_recordings(params=rec_params, tempgen=self.tempgen, verbose=False)

        assert recgen_rs.recordings.shape[0] == num_chan
        assert recgen_rs.recordings.shape[1] == int(duration * fs)
        assert recgen_rs.channel_positions.shape == (num_chan, 3)
        assert len(recgen_rs.spiketrains) == n_neurons
        assert len(recgen_rs.spiketrains) == n_neurons
        assert len(recgen_rs.spiketrains) == n_neurons
        assert len(recgen_rs.spike_traces) == n_neurons
        del recgen_rs

    def test_recordings_backend(self):
        print('Test recording generation - backend')
        ne = 2
        ni = 1
        duration = 3

        rec_params = mr.get_default_recordings_params()

        rec_params['spiketrains']['n_exc'] = ne
        rec_params['spiketrains']['n_inh'] = ni
        rec_params['spiketrains']['duration'] = duration
        n_jitter = 2
        rec_params['templates']['n_jitters'] = n_jitter
        rec_params['recordings']['modulation'] = 'none'

        rec_params['seeds']['templates'] = 0
        rec_params['seeds']['spiketrains'] = 0
        rec_params['seeds']['convolution'] = 0
        rec_params['seeds']['noise'] = 0

        n_jobs = [1, 2]
        chunk_durations = [0, 1]

        for n in n_jobs:
            for ch in chunk_durations:
                print('Test recording backend with', n, 'jobs - chunk', ch)
                rec_params['chunk_duration'] = n

                recgen_memmap = mr.gen_recordings(params=rec_params, tempgen=self.tempgen, tmp_mode='memmap',
                                                  verbose=False, n_jobs=n)
                recgen_np = mr.gen_recordings(params=rec_params, tempgen=self.tempgen, tmp_mode=None, verbose=False,
                                              n_jobs=n)

                assert np.allclose(recgen_np.recordings, np.array(recgen_memmap.recordings), atol=1e-4)
                del recgen_memmap, recgen_np

    def test_recordings_seeds(self):
        print('Test recording generation - seeds')
        ne = 2
        ni = 1
        duration = 3

        rec_params = mr.get_default_recordings_params()

        rec_params['spiketrains']['n_exc'] = ne
        rec_params['spiketrains']['n_inh'] = ni
        rec_params['spiketrains']['duration'] = duration
        n_jitter = 2
        rec_params['templates']['n_jitters'] = n_jitter
        rec_params['recordings']['modulation'] = 'none'

        rec_params['seeds']['templates'] = 0
        rec_params['seeds']['spiketrains'] = 0
        rec_params['seeds']['convolution'] = 0
        rec_params['seeds']['noise'] = 0

        n_jobs = [1, 2]
        chunk_durations = [0, 1]

        for n in n_jobs:
            for ch in chunk_durations:
                print('Test recording backend with', n, 'jobs - chunk', ch)
                rec_params['chunk_duration'] = n

                recgen1 = mr.gen_recordings(params=rec_params, tempgen=self.tempgen, tmp_mode='memmap',
                                            verbose=False, n_jobs=n)
                recgen2 = mr.gen_recordings(params=rec_params, tempgen=self.tempgen, tmp_mode=None, verbose=False,
                                            n_jobs=n)

                assert np.allclose(recgen1.recordings, np.array(recgen2.recordings), atol=1e-4)
                del recgen1, recgen2

    def test_recordings_dtype(self):
        print('Test recording generation - dtype')
        ne = 2
        ni = 1
        duration = 1

        dtypes = ['float16', 'float32', 'float64']
        rec_params = mr.get_default_recordings_params()
        rec_params['spiketrains']['n_exc'] = ne
        rec_params['spiketrains']['n_inh'] = ni
        rec_params['spiketrains']['duration'] = duration
        n_jitter = 3
        rec_params['templates']['n_jitters'] = n_jitter
        rec_params['recordings']['modulation'] = 'none'

        for i, dt in enumerate(dtypes):
            print('Dtype:', dt)
            rec_params['recordings']['dtype'] = dt
            recgen_dt = mr.gen_recordings(params=rec_params, tempgen=self.tempgen, verbose=False)

            assert recgen_dt.recordings[0, 0].dtype == dt
            del recgen_dt

    def test_default_params(self):
        print('Test default params')
        info, info_folder = mr.get_default_config()
        cell_models_folder = info['cell_models_folder']
        tempgen = mr.gen_templates(cell_models_folder, params={'n': 2}, templates_tmp_folder=info['templates_folder'])
        recgen = mr.gen_recordings(templates=self.test_dir + '/templates.h5', verbose=False)
        recgen.params['recordings']['noise_level'] = 0
        recgen.generate_recordings()
        recgen_loaded = mr.load_recordings(self.test_dir + '/recordings.h5', verbose=True)
        recgen_loaded.params['recordings']['noise_level'] = 0
        recgen_loaded.generate_recordings()
        recgen_empty = mr.RecordingGenerator(rec_dict={}, info={})

        n = 2
        num_cells = self.num_cells
        templates_params = self.templates_params

        assert tempgen.templates.shape[0] == (n * num_cells)
        assert len(tempgen.locations) == (n * num_cells)
        assert len(tempgen.rotations) == (n * num_cells)
        assert len(tempgen.celltypes) == (n * num_cells)
        assert len(np.unique(tempgen.celltypes)) == num_cells
        assert np.min(tempgen.locations[:, 0]) > templates_params['xlim'][0] \
               and np.max(tempgen.locations[:, 0]) < templates_params['xlim'][1]

        assert recgen.recordings.shape[0] == self.num_chan
        assert recgen.channel_positions.shape == (self.num_chan, 3)
        assert recgen_loaded.recordings.shape[0] == self.num_chan
        assert recgen_loaded.channel_positions.shape == (self.num_chan, 3)
        assert len(recgen_empty.recordings) == 0
        del recgen, recgen_empty

    def test_cli(self):
        default_config, mearec_home = mr.get_default_config()

        runner = CliRunner()
        result = runner.invoke(cli, ["--help"])
        assert result.exit_code == 0
        result = runner.invoke(cli, ["default-config"])
        assert result.exit_code == 0
        result = runner.invoke(cli, ["available-probes"])
        assert result.exit_code == 0
        result = runner.invoke(cli, ["gen-templates", '-n', '2', '--no-parallel', '--recompile', '-r', '3drot',
                                     '-nc', '2', '-ov', '20', '-s', '1', '-mind', '10', '-maxd', '100',
                                     '-drst', '10', '-v'])
        assert result.exit_code == 0
        result = runner.invoke(cli, ["gen-recordings", '-t', self.test_dir + '/templates.h5', '-ne', '2', '-ni', '1',
                                     '-fe', '5', '-fi', '15', '-se', '1', '-si', '1', '-mr', '0.2',
                                     '-rp', '2', '-p', 'poisson', '-md', '1', '-mina', '10', '-maxa', '1000',
                                     '--fs', '32000', '-sr', '0', '-sj', '1', '-nl', '10', '-m', 'none',
                                     '-chd', '0', '-nseed', '10', '-hd', '30', '-cn', '-cp', '500',
                                     '-cq', '1', '-rnf', '1', '-stseed', '100', '-tseed', '10',
                                     '--filter', '-fc', '500', '-fo', '3', '--overlap', '-ot', '0.8', '--extract-wf',
                                     '-angt', '15', '-drvel', '10', '-tsd', '1', '-v'])
        print(result.output)
        assert result.exit_code == 0
        result = runner.invoke(cli, ["set-templates-params", '.'])
        assert result.exit_code == 0
        result = runner.invoke(cli, ["set-templates-params", default_config['templates_params']])
        assert result.exit_code == 0
        result = runner.invoke(cli, ["set-recordings-params", '.'])
        assert result.exit_code == 0
        result = runner.invoke(cli, ["set-recordings-params", default_config['recordings_params']])
        assert result.exit_code == 0
        result = runner.invoke(cli, ["set-cell-models-folder", './cell_models'])
        assert result.exit_code == 0
        result = runner.invoke(cli, ["set-cell-models-folder", default_config['cell_models_folder']])
        assert result.exit_code == 0
        result = runner.invoke(cli, ["set-templates-folder", '.'])
        assert result.exit_code == 0
        result = runner.invoke(cli, ["set-recordings-folder", '.'])
        assert result.exit_code == 0
        result = runner.invoke(cli, ["set-templates-folder", './templates', '--create'])
        assert result.exit_code == 0
        result = runner.invoke(cli, ["set-recordings-folder", './recordings', '--create'])
        assert result.exit_code == 0
        result = runner.invoke(cli, ["set-templates-folder", default_config['templates_folder']])
        assert result.exit_code == 0
        result = runner.invoke(cli, ["set-recordings-folder", default_config['recordings_folder']])
        assert result.exit_code == 0

    def test_simulate_cell(self):
        cell_folder = mr.get_default_cell_models_folder()
        params = mr.get_default_templates_params()

        target_spikes = [3, 50]
        params['target_spikes'] = target_spikes
        cells = os.listdir(cell_folder)
        cell_name = [c for c in cells if 'TTPC1' in c][0]
        cell_path = os.path.join(cell_folder, cell_name)

        cell, v, i = mr.run_cell_model(cell_model_folder=cell_path, sim_folder=None, verbose=True,
                                       save=False, return_vi=True, **params)
        c = mr.return_bbp_cell_morphology(cell_name, cell_folder)
        assert target_spikes[0] <= len(v) <= target_spikes[1]
        assert target_spikes[0] <= len(i) <= target_spikes[1]
        assert len(c.xmid) == len(c.ymid) and len(c.xmid) == len(c.zmid)


if __name__ == '__main__':
    # ~ unittest.main()

    TestGenerators().setUpClass()
    TestGenerators().test_recordings_backend()



from MEArec.tools import load_templates, load_recordings, save_recording_generator, save_template_generator, \
    get_default_config, plot_templates, plot_recordings, plot_rasters, plot_waveforms, get_templates_features, \
    plot_pca_map, plot_amplitudes, get_default_recordings_params, get_default_templates_params, \
    get_default_cell_models_folder
from MEArec.generation_tools import gen_recordings, gen_templates, gen_spiketrains
from MEArec.generators import RecordingGenerator, SpikeTrainGenerator, TemplateGenerator
from MEArec.simulate_cells import return_bbp_cell, run_cell_model, return_bbp_cell_morphology, \
    calculate_extracellular_potential, calc_extracellular, simulate_templates_one_cell

from .version import version as __version__
